import datetime

from pyplanet.apps.contrib.karma.models import Karma
from pyplanet.apps.contrib.local_records.models import LocalRecord
from pyplanet.apps.core.maniaplanet.models import Player, Map
from pyplanet.apps.core.statistics.models import Score
from pyplanet.contrib.converter.base import BaseConverter


class ManiacontrolConverter(BaseConverter):
	"""
	The Maniacontrol Converter uses MySQL to convert the data to the new PyPlanet structure.

	Please take a look at :doc:`Migrating from other controllers </convert/index>` pages for information on how to use
	this.
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.player_cache = dict()
		self.map_cache = dict()

		if not self.prefix:
			self.prefix = 'mc_'

	async def migrate(self, _):
		print('Migrating players...')
		await self.migrate_players()

		print('Migrating maps...')
		await self.migrate_maps()

		print('Migrating records...')
		await self.migrate_local_records()

		print('Migrating karma...')
		await self.migrate_karma()

	async def migrate_players(self):
		with self.connection.cursor() as cursor:
			cursor.execute('SELECT * FROM {prefix}players'.format(prefix=self.prefix))
			for s_player in cursor.fetchall():
				# Check if we already have this player in our database. If we have, ignore and print message.
				try:
					player = await Player.get_by_login(s_player['login'])
					self.player_cache[player.login] = player

					print('Player with login \'{}\' already exists, skipping..'.format(s_player['login']))
					continue
				except:
					# Not found, create it:
					player = await Player.create(
						login=s_player['login'], nickname=s_player['nickname'],
						last_seen=s_player['changed'],
					)
					self.player_cache[player.login] = player

	async def migrate_maps(self):
		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT * '
				'FROM {prefix}maps '.format(prefix=self.prefix)
			)
			for s_map in cursor.fetchall():
				# Check if we already have this map in our database. If we have, ignore and print message.
				try:
					map_instance = await Map.get_by_uid(s_map['uid'])
					self.map_cache[map_instance.uid] = map_instance

					print('Map with uid \'{}\' already exists, skipping..'.format(s_map['uid']))
					continue
				except:
					# Not found, create it:
					map_instance = await Map.create(
						uid=s_map['uid'], name=s_map['name'], file=s_map['fileName'], author_login=s_map['authorLogin'],
						environment=s_map['environment'], map_type=s_map['mapType'],
					)
					self.map_cache[map_instance.uid] = map_instance

	async def migrate_local_records(self):
		if 'local_records' not in self.instance.apps.apps:
			print('Skipping local records. App not activated!')
			return

		with self.connection.cursor() as cursor:
			try:
				cursor.execute(
					'SELECT 1 FROM {prefix}localrecords'.format(prefix=self.prefix)
				)
				result = cursor.fetchone()
				if not result:
					raise Exception()
			except:
				print('Local records table not found! Skipping...')
				return

		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT record.*, map.uid, player.login '
				'FROM {prefix}localrecords as record, {prefix}maps as map, {prefix}players as player '
				'WHERE record.mapIndex = map.index AND record.playerIndex = player.index '.format(
					prefix=self.prefix
				)
			)
			for s_record in cursor.fetchall():
				# Get map + player
				try:
					map = self.map_cache[s_record['uid']] if s_record['uid'] in self.map_cache else await Map.get(uid=s_record['uid'])
					player = self.player_cache[s_record['login']] if s_record['login'] in self.player_cache else await Player.get(login=s_record['login'])
				except:
					# Skip.
					print('Can\'t convert record, map or player not found. Skipping...')
					continue

				try:
					await LocalRecord.get(map=map, player=player)
					print('Record with uid \'{}\' and player \'{}\' already exists, skipping..'.format(map.uid, player.login))
				except:
					await LocalRecord.create(
						map=map, player=player, score=s_record['time'], checkpoints=s_record['checkpoints'],
						created_at=s_record['changed'], updated_at=datetime.datetime.now()
					)

	async def migrate_karma(self):
		if 'karma' not in self.instance.apps.apps:
			print('Skipping karma. App not activated!')
			return

		with self.connection.cursor() as cursor:
			try:
				cursor.execute(
					'SELECT 1 FROM {prefix}karma'.format(prefix=self.prefix)
				)
				result = cursor.fetchone()
				if not result:
					raise Exception()
			except:
				print('Karma table not found! Skipping...')
				return

		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT rating.*, map.uid, player.login '
				'FROM {prefix}karma AS rating, {prefix}maps AS map, {prefix}players AS player '
				'WHERE rating.mapIndex = map.index AND rating.playerIndex = player.index'.format(
					prefix=self.prefix
				)
			)
			for s_karma in cursor.fetchall():
				# Get map + player
				try:
					map = self.map_cache[s_karma['uid']] if s_karma['uid'] in self.map_cache else await Map.get(uid=s_karma['uid'])
					player = self.player_cache[s_karma['login']] if s_karma['login'] in self.player_cache else await Player.get(login=s_karma['login'])
				except:
					# Skip.
					print('Can\'t convert karma, map or player not found. Skipping...')
					continue

				if s_karma['vote'] == 0:
					continue

				try:
					await Karma.get(map=map, player=player)
					print('Karma with uid \'{}\' and player \'{}\' already exists, skipping..'.format(map.uid, player.login))
				except:
					await Karma.create(
						map=map, player=player, score=-1 if s_karma['vote'] < 0 else 1,
						updated_at=datetime.datetime.now()
					)



"""
Setting app component. Is part of the core contrib component 'setting'.
"""
from pyplanet.apps.core.pyplanet.views.setting import SettingMenuView
from pyplanet.contrib.command import Command


class SettingComponent:
	def __init__(self, app):
		"""
		Initiate setting component.

		:param app: App config instance
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		"""
		self.app = app

	async def on_init(self):
		pass

	async def on_start(self):
		await self.app.instance.permission_manager.register(
			'edit_server_settings', 'Edit server global settings.', app=self.app, min_level=3
		)

		await self.app.instance.command_manager.register(
			# Command('settings', self.player_settings, admin=False),
			Command('settings', self.admin_settings, perms='core.pyplanet:edit_server_settings', admin=True,
					description='Displays and allows editing PyPlanet/plugin settings.'),
		)

	async def player_settings(self, player, *args, **kwargs):
		pass

	async def admin_settings(self, player, *args, **kwargs):
		await SettingMenuView(app=self.app, player=player).display()



"""
Maniaplanet Core Models. This models are used in several apps and should be considered as very stable.
"""
import logging

from peewee import *
from pyplanet.core.db import TimedModel


class Map(TimedModel):
	uid = CharField(
		max_length=50,
		null=False,
		index=True,
		unique=True,
	)
	"""
	The uid of the map will be unique in our database as it's unique over all maps (unique identifier).
	"""

	name = CharField(max_length=150)
	"""
	The name of the map, can contain unparsed styles.
	"""

	file = CharField(max_length=255)
	"""
	Filename and path from root of the maps directory of the dedicated server.
	"""

	author_login = CharField(max_length=100)
	"""
	Author login of the map.
	"""

	author_nickname = CharField(max_length=150, null=True, default=None)
	"""
	Author nickname as of saving the map. (Parsed from the Gbx file).
	"""

	environment = CharField(max_length=30, null=True, default=None)
	"""
	Environment of the map. Sample: Canyon
	"""

	title = CharField(max_length=255, null=True, default=None)
	"""
	The title of the map. Samples: TMCanyon, TMCanyon@nadeo, etc.
	"""

	map_type = CharField(max_length=255, null=True, default=None)
	"""
	Type of map in slashes. Like: Trackmania\Race
	"""

	map_style = CharField(max_length=255, null=True, default=None)
	"""
	Style of the map.
	"""

	num_laps = IntegerField(null=True, default=None)
	"""
	Number of laps if multilap map, else None.
	"""

	num_checkpoints = IntegerField(null=True, default=None)
	"""
	Number of checkpoints. Could be not know (yet).
	"""

	price = IntegerField(null=True, default=None)
	"""
	Price of graphical impact. None if not yet investigated.
	"""

	time_author = IntegerField(null=True, default=None)
	time_bronze = IntegerField(null=True, default=None)
	time_silver = IntegerField(null=True, default=None)
	time_gold = IntegerField(null=True, default=None)
	"""
	Time of author and all medals.
	"""

	mx_id = IntegerField(null=True, default=None, index=True)
	"""
	The MX-id of the map when it was added with PyPlanet. Can be None even if it's been added by PyPlanet or when
	it's been added by another method (manual upload or external software).
	"""

	CACHE = dict()

	def __str__(self):
		return '\'{}\' by {} ({})'.format(self.name, self.author_login, self.uid)

	async def save(self, *args, **kwargs):
		await super().save(*args, **kwargs)
		self.CACHE[self.uid] = self

	async def get_author(self):
		from .player import Player

		if self.author_login:
			await Player.get(login=self.author_login)
		return None

	@classmethod
	async def get_by_uid(cls, uid):
		"""
		Get map by UID.

		:param uid: UId.
		:return: Map instance
		:rtype: pyplanet.apps.core.maniaplanet.models.map.Map
		"""
		if uid in cls.CACHE:
			return cls.CACHE[uid]
		return await cls.get(uid=uid)

	@classmethod
	async def get_or_create_from_info(cls, uid, file, name, author_login, **kwargs):
		"""
		This method will be called from the core, getting or creating a map instance from the information we got from
		the dedicated server.

		:param uid: Map UID
		:param file: Filename
		:param name: Name of map
		:param author_login: Author login
		:param kwargs: Other key arguments, matching the model columns!
		:return: Map instance.
		"""
		needs_save = False

		# HACK: Due to a limited map name length of 150 chars, we want to strip it to the maximum possible.
		# This is a temporary fix and should be better handled in the future.
		if len(name) > 150:
			name = name[:150]
			logging.getLogger(__name__).warning('Map name is very long, truncating to 150 chars.')

		try:
			map = await cls.get_by_uid(uid)
			if map.file != file or map.name != name:
				map.file = file
				map.name = name
				needs_save = True
			if 'mx_id' in kwargs and map.mx_id != kwargs['mx_id']:
				map.mx_id = kwargs['mx_id']
				needs_save = True
		except DoesNotExist:
			map = Map(uid=uid, file=file, name=name, author_login=author_login)
			needs_save = True

		# Update from the kwargs.
		for k, v in kwargs.items():
			if v is not None and hasattr(map, k) and getattr(map, k) != v:
				setattr(map, k, v)
				needs_save = True

		if needs_save:
			await map.save()

		return map



from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet.callbacks.player import player_enter_player_slot, player_enter_spectator_slot
from pyplanet.apps.core.maniaplanet.callbacks.other import server_password


class ManiaplanetConfig(AppConfig):
	name = 'pyplanet.apps.core.maniaplanet'
	core = True

	async def on_start(self):
		self.context.signals.register_signal(player_enter_player_slot)
		self.context.signals.register_signal(player_enter_spectator_slot)
		self.context.signals.register_signal(server_password)



from pyplanet.views.generics.list import ManualListView


class MusicListView(ManualListView):
	title = 'Songs'
	icon_style = 'Icons128*128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.provide_search = True

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'index',
				'sorting': True,
				'searching': False,
				'width': 10,
				'type': 'label'
			},
			{
				'name': 'Song',
				'index': 'song_name',
				'sorting': True,
				'searching': True,
				'width': 100,
				'action': self.action_playlist
			},
			{
				'name': 'Artist',
				'index': 'song_artist',
				'sorting': True,
				'searching': True,
				'width': 100,
			},
		]

	async def get_data(self):
		items = []
		song_list = self.app.songs
		for song in song_list:
			tags = song[1]
			items.append({
				'index': song_list.index(song) + 1,
				'song_name': tags.get('title', '-unknown title-'),
				'song_artist': tags.get('artist', '-unknown artist-'),
			})
		return items

	async def action_playlist(self, player, values, song_info, *args, **kwargs):
		await self.app.add_to_playlist(player, song_info['index']-1)


class PlaylistView(ManualListView):
	title = 'Playlist'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.provide_search = True#

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'index',
				'sorting': True,
				'searching': False,
				'width': 10,
				'type': 'label'
			},
			{
				'name': 'Song',
				'index': 'song_name',
				'sorting': True,
				'searching': True,
				'width': 100,
				'action': self.action_drop
			},
			{
				'name': 'Artist',
				'index': 'song_artist',
				'sorting': True,
				'searching': True,
				'width': 50,
			},
			{
				'name': 'Requested by',
				'index': 'juke_player',
				'sorting': True,
				'searching': True,
				'width': 60
			}
		]

	async def action_drop(self, player, values, song_info, **kwargs):
		await self.app.drop_from_playlist(player, song_info)

	async def get_data(self):
		items = []
		playlist = self.app.playlist
		for song in playlist:
			tags = song['song'][1]
			items.append({
				'index': playlist.index(song) + 1,
				'song_name': tags.get('title'),
				'song_artist': tags.get('artist'),
				'juke_player': song['player']
			})
		return items



import asynctest
from jinja2 import Template

from pyplanet.core import Controller
from pyplanet.core.ui.template import load_template


class TestTemplate(asynctest.TestCase):
	async def test_template_loading(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		template = await load_template('core.views/generics/list.xml')
		assert template and template.template
		assert isinstance(template.template, Template)

	async def test_template_rendering(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		template = await load_template('core.views/generics/list.xml')
		body = await template.render(
			title='TRY_TO_SEARCH_THIS'
		)
		assert 'TRY_TO_SEARCH_THIS' in body



from pyplanet.views.generics import ManualListView
from .base import StatsView


class TopRecordsView(StatsView):
	template_name = 'core.statistics/menu.xml'

	async def get_context_data(self):
		context = await super().get_context_data()
		context['options'] = [
			dict(name='Top Records', view='pyplanet.apps.core.statistics.views.records.TopRecords'),
		]


class TopSumsView(ManualListView):
	title = 'Top players on the server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player, topsums):
		"""
		Init topsums list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False
		self.topsums = topsums

	async def get_data(self):
		data = list()
		for idx, (player, (first, second, third)) in enumerate(self.topsums):
			data.append(dict(
				place=idx+1,
				player_nickname=player.nickname,
				first=first,
				second=second,
				third=third,
				total=first + second + third,
			))

		return data

	async def destroy(self):
		self.topsums = None
		await super().destroy()

	def destroy_sync(self):
		self.topsums = None
		super().destroy_sync()

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'place',
				'sorting': False,
				'searching': False,
				'width': 6,
				'type': 'label'
			},
			{
				'name': 'Player',
				'index': 'player_nickname',
				'sorting': False,
				'searching': False,
				'width': 60,
				'type': 'label'
			},
			{
				'name': '1st local',
				'index': 'first',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
			{
				'name': '2nd local',
				'index': 'second',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
			{
				'name': '3rd local',
				'index': 'third',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
			{
				'name': 'total top 3',
				'index': 'total',
				'sorting': True,
				'searching': False,
				'width': 30,
				'type': 'label'
			},
		]



import asyncio
import logging

from xmlrpc.client import Fault

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core.ui.ui_properties import UIProperties
from pyplanet.utils.log import handle_exception

logger = logging.getLogger(__name__)


class _BaseUIManager:
	def __init__(self, instance):
		"""
		Initiate manager.

		:param instance: Instance of controller.
		:type instance: pyplanet.core.instance.Instance
		"""
		self.instance = instance
		self.manialinks = dict()
		self.send_queue = list()

	async def on_start(self):
		asyncio.ensure_future(self.send_loop())

	async def send_loop(self):
		while True:
			await asyncio.sleep(0.25)
			if len(self.send_queue) == 0:
				continue

			# Copy send queue and clear the global one
			queue = self.send_queue.copy()
			self.send_queue.clear()

			# Process and push out the queue.
			try:
				await self.instance.gbx.multicall(*queue)
			except Fault as e:
				if 'Login unknown' in str(e):
					return
				logger.exception(e)
				handle_exception(exception=e, module_name=__name__, func_name='send_loop')
			except Exception as e:
				logger.exception(e)
				handle_exception(exception=e, module_name=__name__, func_name='send_loop')

	def get_manialink_by_id(self, identifier):
		"""
		Get Manialink instance by ManiaLink identifier.

		:param identifier: Identifier string
		:type identifier: str
		:return: ManiaLink instance or None
		:rtype: pyplanet.core.ui.components.manialink._ManiaLink
		"""
		if identifier in self.manialinks:
			return self.manialinks[identifier]
		return None

	async def send(self, manialink, players=None, **kwargs):
		"""
		Send manialink to player(s).

		:param manialink: ManiaLink instance.
		:param players: Player instances or logins to post to. None to globally send.
		:type manialink: pyplanet.core.ui.components.manialink._ManiaLink
		"""
		queries = list()
		if isinstance(players, list):
			for_logins = [p.login if isinstance(p, Player) else p for p in players]
		elif manialink.player_data:
			for_logins = list(manialink.player_data.keys())
		else:
			for_logins = list()

		# Register to the manialink context.
		if manialink.id not in self.manialinks:
			self.manialinks[manialink.id] = manialink

		is_global = await manialink.is_global()
		if not is_global:
			for login in for_logins:
				if login not in manialink.player_data:
					continue

				if await manialink.get_template() and not manialink.body:
					body = await manialink.render(player_login=login)
				elif manialink.body:
					body = manialink.body
				else:
					raise Exception('Manialink has no body or template defined!')

				# Add manialink tag to body.
				body = '<manialink version="{}" id="{}">{}</manialink>'.format(manialink.version, manialink.id, body)

				# Prepare query
				queries.append(self.instance.gbx(
					'SendDisplayManialinkPageToLogin', login, body, manialink.timeout, manialink.hide_click
				))

		else:
			# Render/body
			if await manialink.get_template() and not manialink.body:
				body = await manialink.render()
			elif manialink.body:
				body = manialink.body
			else:
				raise Exception('Manialink has no body or template defined!')

			# Add manialink tag to body.
			body = '<manialink version="{}" id="{}">{}</manialink>'.format(manialink.version, manialink.id, body)

			# Add normal queries.
			if for_logins and len(for_logins) > 0:
				for login in for_logins:
					# Prepare query
					queries.append(self.instance.gbx(
						'SendDisplayManialinkPageToLogin', login, body, manialink.timeout, manialink.hide_click
					))
			else:
				# Prepare query
				queries.append(self.instance.gbx(
					'SendDisplayManialinkPage', body, manialink.timeout, manialink.hide_click
				))

		# Hide ALT menus (shootmania).
		if self.instance.game.game == 'sm' and manialink.disable_alt_menu:
			if is_global:
				queries.extend([
					self.instance.gbx('Maniaplanet.UI.SetAltScoresTableVisibility', player.login, 'false', encode_json=False, response_id=False)
					for player in self.instance.player_manager.online
				])
			else:
				queries.extend([
					self.instance.gbx('Maniaplanet.UI.SetAltScoresTableVisibility', login, 'false', encode_json=False, response_id=False)
					for login in for_logins
				])

		# It the manialink wants rate limitting with the relaxed updating feature (mostly used for widgets), add to send queue
		if getattr(manialink, 'relaxed_updating', False):
			self.send_queue.extend(queries)
			return

		# Execute calls, ignore login unknown (player just left).
		try:
			await self.instance.gbx.multicall(*queries)
		except Fault as e:
			if 'Login unknown' in str(e):
				return
			raise

	async def hide(self, manialink, logins=None):
		"""
		Send manialink to player(s).

		:param manialink: ManiaLink instance.
		:param logins: Logins to post to. None to globally send.
		:type manialink: pyplanet.core.ui.components.manialink._ManiaLink
		"""
		body = '<manialink id="{}"></manialink>'.format(manialink.id)
		queries = list()
		if logins and len(logins) > 0:
			queries.append(
				self.instance.gbx('SendDisplayManialinkPageToLogin', ','.join(logins), body, 0, False)
			)

			# Show alt menu again.
			if self.instance.game.game == 'sm' and manialink.disable_alt_menu:
				queries.extend([
					self.instance.gbx('Maniaplanet.UI.SetAltScoresTableVisibility', login, 'true', encode_json=False, response_id=False)
					for login in logins
				])
		else:
			queries.append(self.instance.gbx('SendDisplayManialinkPage', body, 0, False))
			if self.instance.game.game == 'sm' and manialink.disable_alt_menu:
				queries.extend([
					self.instance.gbx('Maniaplanet.UI.SetAltScoresTableVisibility', player.login, 'true', encode_json=False, response_id=False)
					for player in self.instance.player_manager.online
				])

		# It the manialink wants rate limitting with the relaxed updating feature (mostly used for widgets), add to send queue
		if getattr(manialink, 'relaxed_updating', False):
			self.send_queue.extend(queries)
			return

		# Execute queries.
		await self.instance.gbx.multicall(*queries)

	async def destroy(self, manialink, logins=None):
		if manialink.id in self.manialinks:
			del self.manialinks[manialink.id]
		return await self.hide(manialink, logins)


class GlobalUIManager(_BaseUIManager):
	def __init__(self, instance):
		super().__init__(instance)
		self.app_managers = dict()
		self.properties = UIProperties(self.instance)

	async def on_start(self):
		await super().on_start()
		await self.properties.on_start()

		# Start app ui managers.
		await asyncio.gather(*[
			m.on_start() for m in self.app_managers.values()
		])

	def get_manialink_by_id(self, identifier):
		"""
		Get Manialink instance by ManiaLink identifier. (From all apps ui managers as well).

		:param identifier: Identifier string
		:type identifier: str
		:return: ManiaLink instance or None
		:rtype: pyplanet.core.ui.components.manialink._ManiaLink
		"""
		global_ui = super().get_manialink_by_id(identifier)
		if global_ui is not None:
			return global_ui

		for app_manager in self.app_managers.values():
			app_manialink = app_manager.get_manialink_by_id(identifier)
			if app_manialink is not None:
				return app_manialink

	def create_app_manager(self, app_config):
		"""
		Create app ui manager.

		:param app_config: App Config instance.
		:type app_config: pyplanet.apps.config.AppConfig
		:return: UI Manager
		:rtype: pyplanet.core.ui.AppUIManager
		"""
		if app_config.label not in self.app_managers:
			self.app_managers[app_config.label] = AppUIManager(self.instance, app_config)
		return self.app_managers[app_config.label]


class AppUIManager(_BaseUIManager):
	"""
	The App UI manager is here to maintain the context of the app and have it destroy all the listeners when the app
	is unloaded.
	"""

	def __init__(self, instance, app):
		"""
		Initiate app ui manager.

		:param instance: Controller instance.
		:param app: App Config instance.
		:type instance: pyplanet.core.instance.Instance
		:type app: pyplanet.apps.config.AppConfig
		"""
		super().__init__(instance)
		self.app = app

	async def on_destroy(self):
		links = self.manialinks.copy()
		for ml in links.values():
			try:
				await ml.destroy()
			except Exception as e:
				logger.warning('Got exception while destroying apps UI: {}'.format(str(e)))
				logger.debug(e)
			try:
				del ml
			except:
				pass

		self.manialinks.clear()



from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core.db import TimedModel

from peewee import *


class TestModel(TimedModel):
	player = ForeignKeyField(Player)
	test = BooleanField(default=True)
	sample_field = CharField(default='unknown', db_column='sample')



from pyplanet.utils import times
from pyplanet.views.generics.list import ManualListView
from pyplanet.views.generics.widget import TimesWidgetView


class BestCpTimesWidget(TimesWidgetView):
	widget_x = -124.75
	widget_y = 90
	z_index = 130
	size_x = 250
	size_y = 18
	title = 'Best CPs'

	template_name = 'best_cps/widget_top.xml'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widget_bestcps'
		self.action = self.action_cptimeslist

	async def get_player_data(self):
		data = await super().get_player_data()
		cps = {}
		for idx, player in enumerate(self.app.instance.player_manager.online):
			list_cps = []
			for pcp in self.app.best_cp_times:
				list_time = {
					'index': pcp.cp,
					'color': "$0f3" if player.login == pcp.player.login else "$ff0",
					'cptime': times.format_time(pcp.time),
					'nickname': pcp.player.nickname,
					'login': pcp.player.login
				}
				list_cps.append(list_time)
			cps[player.login] = {'cps': list_cps}

		data.update(cps)
		return data

	async def action_cptimeslist(self, player, **kwargs):
		view = CpTimesListView(self.app)
		await view.display(player=player.login)
		return view

	# TODO: remove/rework when finishing widget for rounds/cup/team mode gets reworked
	# Only show the widget in TimeAttack mode as it interferes with UI elements in the other modes
	async def display(self, player=None, **kwargs):
		await super().display()
		# current_script = await self.app.instance.mode_manager.get_current_script()
		# if 'TimeAttack' in current_script:
		#
		# else:
		# 	for idx, player in enumerate(self.app.instance.player_manager.online):
		# 		await super().close(player)


class CpTimesListView(ManualListView):
	title = 'Best CP times in current round'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	fields = [
		{
			'name': '#',
			'index': 'index',
			'sorting': True,
			'searching': False,
			'width': 10,
			'type': 'label'
		},
		{
			'name': 'Player',
			'index': 'player_nickname',
			'sorting': False,
			'searching': True,
			'width': 70
		},
		{
			'name': 'Time',
			'index': 'record_time',
			'sorting': True,
			'searching': False,
			'width': 30,
			'type': 'label'
		},
	]

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.provide_search = False

	async def get_data(self):
		items = []
		list_times = self.app.best_cp_times
		for pcp in list_times:
			items.append({
				'index': pcp.cp,
				'player_nickname': pcp.player.nickname,
				'record_time': times.format_time(pcp.time)
			})

		return items



import asynctest

from pyplanet.core.instance import Controller


class TestInstanceInit(asynctest.TestCase):
	async def test_gbx_init(self):
		instance = Controller.prepare(name='default').instance
		await instance.gbx.connect()
		self.assertGreater(len(instance.gbx.gbx_methods), 0)
		await instance.gbx.disconnect()
		del instance

	async def test_startup(self):
		instance = Controller.prepare(name='default').instance
		await instance._start()
		self.assertGreater(len(instance.gbx.gbx_methods), 0)
		await instance.gbx.disconnect()
		try:
			await instance.db.engine.disconnect()
		except:
			pass
		del instance



import importlib
import os

from pyplanet.conf.backends.base import ConfigBackend
from pyplanet.core.exceptions import ImproperlyConfigured


class PythonConfigBackend(ConfigBackend):
	name = 'python'

	def __init__(self, **options):
		super().__init__(**options)

		self.module = None

	def load(self):
		# Make sure we load the defaults first.
		super().load()

		# Prepare the loading.
		self.module = os.environ.get('PYPLANET_SETTINGS_MODULE', 'settings')

		if not self.module:
			raise ImproperlyConfigured(
				'Settings module is not defined! Please define PYPLANET_SETTINGS_MODULE in your environment or start script.'
			)

		# Add the module itself to the configuration.
		self.settings['SETTINGS_MODULE'] = self.module

		# Load the module, put the settings into the local context.
		try:
			module = importlib.import_module(self.module)
		except ModuleNotFoundError as e:
			raise ImproperlyConfigured(
				'The settings module doesn\'t contain any submodules or files to load! Please make sure '
				'your settings module exist or contains the files base.py and apps.py. Your module: {}'.format(self.module)
			) from e

		# Load from the modules.
		processed = 0
		for setting in dir(module):
			if setting.isupper():
				self.settings[setting] = getattr(module, setting)
				processed += 1

		# Check for empty results.
		if processed < 1:
			raise ImproperlyConfigured(
				'The settings module doesn\'t contain any submodules or files to load! Please make sure '
				'your settings module exist or contains the files base.py and apps.py. Your module: {}'.format(self.module)
			)



"""
Server Admin methods and functions.
"""
import asyncio

from pyplanet.apps.core.maniaplanet.callbacks.player import player_chat
from pyplanet.contrib.command import Command
from xmlrpc.client import Fault

from pyplanet.apps.contrib.admin.views.setting import ModeSettingMenuView


class ServerAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance

		self.chat_redirection = False

	async def on_start(self):
		await self.instance.permission_manager.register('callvoting', 'Handle server callvoting', app=self.app, min_level=1)
		await self.instance.permission_manager.register('password', 'Set the server passwords', app=self.app, min_level=2)
		await self.instance.permission_manager.register('servername', 'Set the server name', app=self.app, min_level=2)
		await self.instance.permission_manager.register('mode', 'Set the server game mode', app=self.app, min_level=2)
		await self.instance.permission_manager.register('chat_toggle', 'Turn the public chat on or off', app=self.app, min_level=2)

		await self.instance.command_manager.register(
			Command(command='cancelcallvote', aliases=['cancelcall'], target=self.cancel_callvote, perms='admin:callvoting',
					admin=True, description='Cancels the current callvote (server vote).'),
			Command(command='setpassword', aliases=['srvpass'], target=self.set_password, perms='admin:password', admin=True,
					description='Sets the player password of the server.').add_param(name='password', required=False),
			Command(command='setspecpassword', aliases=['spectpass'], target=self.set_spec_password, perms='admin:password',
					admin=True, description='Sets the spectator password of the server.').add_param(name='password', required=False),
			Command(command='servername', target=self.set_servername, perms='admin:servername', admin=True,
					description='Changes the name of the server.').add_param(name='server_name', required=True, nargs='*'),
			Command(command='mode', target=self.set_mode, perms='admin:mode', admin=True,
					description='Changes the gamemode of the server.').add_param(name='mode', required=True, nargs='*'),
			Command(command='modesettings', target=self.mode_settings, perms='admin:mode', admin=True,
					description='Displays and allows for updating of the gamemode settings.')
				.add_param(name='setting', required=False)
				.add_param(name='content', required=False),
			Command(command='chat', target=self.chat_toggle, perms='admin:chat_toggle', admin=True,
					description='Enables/disables the chat.')
				.add_param(name='enable', required=False),
		)

		# Register signal receivers.
		player_chat.register(self.on_chat)

	async def cancel_callvote(self, player, data, **kwargs):
		current_callvote = await self.instance.gbx('GetCurrentCallVote')

		if current_callvote['CmdName'] == '':
			message = '$i$f00There is currently no callvote active.'
			await self.instance.chat(message, player)
		else:
			message = '$ff0Admin $fff{}$z$s$ff0 has cancelled the current call vote.'.format(player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('CancelVote'),
				self.instance.chat(message)
			)

	async def chat_toggle(self, player, data, **kwargs):
		if data.enable:
			self.chat_redirection = bool(data.enable.lower() == 'on')
		else:
			self.chat_redirection = bool(not self.chat_redirection)
		await self.instance.gbx('ChatEnableManualRouting', self.chat_redirection, True)
		await self.instance.chat('$ff0Admin $fff{}$z$s$ff0 {} public chat'.format(
			player.nickname, 'disabled' if self.chat_redirection else 'enabled'
		))

	async def on_chat(self, player, text, cmd, **kwargs):
		if not cmd and self.chat_redirection:
			if player.level > 0:
				asyncio.ensure_future(self.instance.chat(
					'$z[{}$z$s] {}'.format(player.nickname, text), raw=True
				))

	async def set_mode(self, player, data, **kwargs):
		mode = (' '.join(data.mode))
		lower_mode = mode.lower()
		if self.instance.game.game == 'tm':
		
			if lower_mode == 'ta' or lower_mode == 'timeattack':
				mode = 'TimeAttack.Script.txt'
			elif lower_mode == 'laps':
				mode = 'Laps.Script.txt'
			elif lower_mode == 'rounds':
				mode = 'Rounds.Script.txt'
			elif lower_mode == 'cup':
				mode = 'Cup.Script.txt'
			elif lower_mode == 'chase':
				mode = 'Chase.Script.txt'
			elif lower_mode == 'team':
				mode = 'Team.Script.txt'
			
		if self.instance.game.game == 'tmnext':
		
			if lower_mode == 'ta' or lower_mode == 'timeattack':
				mode = 'Trackmania/TM_TimeAttack_Online.Script.txt'
			elif lower_mode == 'laps':
				mode = 'Trackmania/TM_Laps_Online.Script.txt'
			elif lower_mode == 'rounds':
				mode = 'Trackmania/TM_Rounds_Online.Script.txt'
			elif lower_mode == 'cup':
				mode = 'Trackmania/TM_Cup_Online.Script.txt'
			elif lower_mode == 'team':
				mode = 'Trackmania/TM_Teams_Online.Script.txt'
			elif lower_mode == 'knockout':
				mode = 'Trackmania/TM_Knockout_Online.Script.txt'
			elif lower_mode == 'champion':
				mode = 'Trackmania/TM_Champion_Online.Script.txt'

		try:
			await self.instance.mode_manager.set_next_script(mode)
		except Exception as e:
			message = '$ff0Mode change failed: {}'.format(str(e))
			await self.instance.chat(message, player)
			return
		message = '$ff0Admin $fff{}$z$s$ff0 has changed the next mode to {}'.format(
			player.nickname, mode
		)
		await self.instance.chat(message)

	async def mode_settings(self, player, data, **kwargs):
		setting_name = data.setting
		if setting_name is None:
			view = ModeSettingMenuView(self.app, player)
			await view.display(player=player.login)
		else:
			if not data.content:
				message = '$i$f00Setting a mode setting requires $fff2$f00 parameters.'
				await self.instance.chat(message, player)
				return

			current_settings = await self.instance.mode_manager.get_settings()
			setting_value = data.content
			if setting_name not in current_settings:
				message = '$i$f00Unknown mode setting "$fff{}$f00".'.format(setting_name)
				await self.instance.chat(message, player)
				return

			current_value = current_settings[setting_name]
			current_type = type(current_value)

			type_setting = None
			try:
				if isinstance(current_value, bool):
					lower_setting_value = setting_value.lower()
					if lower_setting_value == 'true' or setting_value == '1':
						type_setting = True
					elif lower_setting_value == 'false' or setting_value == '0':
						type_setting = False
					else:
						raise ValueError
				else:
					type_setting = current_type(setting_value)

				type_setting = current_type(setting_value)
				await self.instance.mode_manager.update_settings({
					setting_name: type_setting
				})

				message = '$ff0Changed mode setting "$fff{}$ff0" to "$fff{}$ff0" (was: "$fff{}$ff0").'.format(setting_name, type_setting, current_value)
				await self.instance.chat(message, player)
			except ValueError:
				message = '$i$f00Unable to cast "$fff{}$f00" to required type ($fff{}$f00) for "$fff{}$f00".'.format(setting_value, current_type, setting_name)
				await self.instance.chat(message, player)
			except Fault as exception:
				message = '$i$f00Unable to set "$fff{}$f00" to "$fff{}$f00": $fff{}$f00.'.format(setting_name, type_setting, exception)
				await self.instance.chat(message, player)

	async def set_servername(self, player, data, **kwargs):
		name = ' '.join(data.server_name)
		message = '$ff0Admin $fff{}$z$s$ff0 has changed the server name into {}'.format(player.nickname, name)
		await self.instance.gbx.multicall(
			self.instance.gbx('SetServerName', name),
			self.instance.chat(message)
		)

	async def set_spec_password(self, player, data, **kwargs):
		if data.password is None or data.password == 'none':
			message = '$ff0You removed the spectator password.'
			await self.instance.gbx.multicall(
				self.instance.gbx('SetServerPasswordForSpectator', ''),
				self.instance.chat(message, player)
			)
		else:
			message = '$ff0You changed the spectator password to: "$fff{}$ff0".'.format(data.password)
			await self.instance.gbx.multicall(
				self.instance.gbx('SetServerPasswordForSpectator', data.password),
				self.instance.chat(message, player)
			)

		await self.app.context.signals.get_signal('maniaplanet:server_password').send(
			dict(password=data.password, kind='spectator'), True
		)

	async def set_password(self, player, data, **kwargs):
		if data.password is None or data.password == 'none':
			message = '$ff0You removed the server password.'
			await self.instance.gbx.multicall(
				self.instance.gbx('SetServerPassword', ''),
				self.instance.chat(message, player)
			)
		else:
			message = '$ff0You changed the server password to: "$fff{}$ff0".'.format(data.password)
			await self.instance.gbx.multicall(
				self.instance.gbx('SetServerPassword', data.password),
				self.instance.chat(message, player)
			)

		await self.app.context.signals.get_signal('maniaplanet:server_password').send(
			dict(password=data.password, kind='player'), True
		)



import os
import time
import logging
import multiprocessing

from logging.handlers import QueueListener

from pyplanet.utils.livereload import LiveReload
from pyplanet.god import process

logger = logging.getLogger(__name__)


class EnvironmentPool:
	"""
	This class manages the pool instances for the current environment/installation.

	.. warning::

		You should not have to use this class in any moment!

	"""

	def __init__(self, pool_names, max_restarts=0, options=None):
		self.names = pool_names
		self.queue = multiprocessing.Queue()
		self.pool = dict()
		self.max_restarts = max_restarts
		self.options = options or dict()

		self.dog_path = os.curdir
		self.dog_handler = LiveReload(self)
		# self.dog_observer = Observer()
		# self.dog_observer.schedule(self.dog_handler, self.dog_path, recursive=True)

		if multiprocessing.get_start_method() != 'fork':  # pragma: no cover
			root_logger = logging.getLogger()
			self.log_listener = QueueListener(self.queue, *root_logger.handlers)

		# TODO: Find out how to get the watchdog + livereload working on a later moment.
		# self.dog_observer.start()

		self._restarts = dict()

	@property
	def num_online(self):
		count = 0
		for proc in self.pool.values():
			if not proc.did_die:
				count += 1
		return count

	def populate(self):
		"""
		Populate the pool instance processes, (prepares the processes).
		"""
		for name in self.names:
			self.pool[name] = process.InstanceProcess(queue=self.queue, environment_name=name, options=self.options)
			self._restarts[name] = 0
		return self

	def start(self):
		"""
		Start all processes.
		"""
		for name, proc in self.pool.items():
			proc.start()

	def shutdown(self):
		"""
		Shutdown all processes.
		"""
		for name, proc in self.pool.items():
			logger.info('Shutting down {}...'.format(name))
			if 'detach' in self.options and self.options['detach']:
				proc.shutdown()
			else:
				proc.graceful()
		# self.dog_observer.stop()

	def restart(self, name=None):
		"""
		Restart single process, or all if no name is given.

		:param name: Name or none for all pools.
		"""
		if name:
			self.pool[name] = process.InstanceProcess(queue=self.queue, environment_name=name)
			self._restarts[name] += 1
			self.pool[name].start()
		else:
			for name in self.pool.keys():
				self.restart(name)

	def watchdog(self):
		"""
		Watch all the processes. (Blocking method!).
		"""
		logger.debug('Starting watchdog... watching {} instances'.format(len(self.pool)))

		while True:
			num_alive = 0
			for name, proc in self.pool.items():
				if proc.did_die:
					# Process wants a restart! = exit code 50.
					if proc.exitcode == 50:
						self.restart(name)
						# Make sure we don't kill our god at the end of the loop.
						num_alive = 1
					# Status changed from 'online' to 'offline'
					elif self._restarts[name] < self.max_restarts:
						logger.critical('The instance \'{}\' just died. We will restart the instance!'.format(name))
						self.restart(name)
						num_alive += 1
					else:
						logger.critical('The instance \'{}\' just died. We will not restart!'.format(name))
				else:
					num_alive += 1

			# Check if there are still processes alive.
			if num_alive == 0:
				logger.critical('All instances died. Quitting now...')
				exit(1)

			time.sleep(2)



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models.player import Player


def upgrade(migrator: SchemaMigrator):
	total_playtime = IntegerField(default=0, null=False)
	total_donations = IntegerField(default=0, null=False)

	migrate(
		migrator.add_column(Player._meta.db_table, 'total_playtime', total_playtime),
		migrator.add_column(Player._meta.db_table, 'total_donations', total_donations),
	)


def downgrade(migrator: SchemaMigrator):
	migrate(
		migrator.drop_column(Player._meta.db_table, 'total_playtime'),
		migrator.drop_column(Player._meta.db_table, 'total_donations'),
	)



import collections

from xmlrpc.client import Fault

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.contrib.chat.exceptions import ChatException
from pyplanet.core.gbx.query import Query


class ChatQuery(Query):
	"""
	The chat query is the chat message building class in PyPlanet.

	Please get a new instance from the chat manager with ``instance.chat.prepare()`` and chain your methods from there.
	"""

	def __init__(self, chat_manager, message=None, logins=None, auto_prefix=True):
		"""
		Build a chat query with this class, but please use it with the chat contrib ``prepare`` method.

		:param chat_manager: Chat manager instance.
		:param message: Optional predefined message.
		:param logins: Optional list of logins, or None (default) for global.
		:param auto_prefix: Optional: Automatically add prefix if it's public or private to the message.
		:type chat_manager: pyplanet.contrib.chat.manager.ChatManager
		"""
		self.chat_manager = chat_manager
		self.instance = chat_manager.instance

		self.auto_prefix = auto_prefix

		self._message = message or ''
		self._logins = logins

		query = self.gbx_query
		super().__init__(self.instance.gbx, query.method, *query.args)

	@property
	def method(self):
		return self.gbx_query.method

	@method.setter
	def method(self, _):
		pass

	@property
	def args(self):
		return self.gbx_query.args

	@args.setter
	def args(self, _):
		pass

	def to_players(self, *players):
		"""
		Set the destination of the chat message.

		:param players: Player instance(s) or player login string(s). Can be a list, or a single entry.
		:return: Self reference.
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		# Unpack list in unpacked list if given.
		if len(players) == 1 and isinstance(players[0], collections.Iterable):
			players = players[0]

		# Replace logins.
		if isinstance(players, Player):
			self._logins = set()
			self._logins.add(players.login)
		elif isinstance(players, str):
			self._logins = set()
			self._logins.add(players)
		elif isinstance(players, collections.Iterable) and isinstance(players, collections.Sized):
			self._logins = set()
			self.add_to(players)
		return self

	def add_to(self, *players):
		"""
		Add new recipient to the to list.

		:param players: Player login string(s) or player instance(s).
		:return: Self reference.
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		# Unpack list in unpacked list if given.
		if len(players) == 1 and isinstance(players[0], collections.Iterable):
			players = players[0]

		# Check if we already have login lists.
		if not isinstance(self._logins, set):
			self._logins = set()

		for obj in players:
			if isinstance(obj, Player):
				self._logins.add(obj.login)
			elif isinstance(obj, str):
				self._logins.add(obj)
		return self

	def to_all(self):
		"""
		Send message to all players on server (default).

		:return: Self reference.
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		self._logins = None
		return self

	def message(self, message: str):
		"""
		Set the message payload.

		:param message: Message of the chat message.
		:return: Self reference.
		:rtype: pyplanet.contrib.chat.query.ChatQuery
		"""
		self._message = message
		return self

	def get_formatted_message(self):
		"""
		Get the formatted message. (will get the message string with prefix if applied).

		:return: String
		:rtype: str
		"""
		if not isinstance(self._message, str):
			raise ChatException('Chat message must be defined as a string! Use the .message() on your chat query!')

		message = ''

		# Add prefixes if public or private chat message.
		if self.auto_prefix:
			if isinstance(self._logins, set):
				message = '$z$s$fff» '
			else:
				message = '$z$s$fff»» '

		# Add the message payload.
		return message + self._message

	def prepare(self):
		"""
		Get a prepared gbx query for this chat message.

		:return: Prepared GBX query.
		:rtype: pyplanet.core.gbx.query.Query
		"""
		super().prepare()
		return self.gbx_query

	@property
	def gbx_query(self):
		"""
		Get a prepared gbx query for this chat message.

		:return: Prepared GBX query.
		:rtype: pyplanet.core.gbx.query.Query
		"""
		method = 'ChatSendServerMessage'
		args = list()
		args.append(self.get_formatted_message())

		if isinstance(self._logins, set):
			method = 'ChatSendServerMessageToLogin'
			args.append(','.join(self._logins))

		return self.instance.gbx(method, *args)

	async def execute(self):  # pragma: no cover
		"""
		Execute the chat message sending query. Please don't use this when you send multiple chat messages or actions!

		:return: Result of query.
		"""
		try:
			return await self.gbx_query.execute()
		except Fault as e:
			if 'Login unknown' in e.faultString:
				return True  # Ignore
			raise




class StorageInterface:
	async def open(self, file: str, mode: str = 'rb'):
		raise NotImplementedError

	async def open_map(self, file: str, mode: str = 'rb'):
		raise NotImplementedError

	async def open_match_settings(self, file: str, mode: str = 'r'):
		raise NotImplementedError


class StorageDriver:

	def __init__(self, instance, config: dict = None):
		"""
		Initiate storage driver.
		
		:param instance: Instance instance :P.
		:param config: Driver configuration.
		:type instance: pyplanet.core.instance.Instance
		:type config: dict
		"""
		self.instance = instance
		self.config = config or {}

		self.map_dir = None
		self.skin_dir = None
		self.data_dir = None
		self.base_dir = None

	def absolute(self, path):
		return '{}{}{}'.format(self.base_dir or '', '/' if self.base_dir else '', path)

	async def __aenter__(self, **kwargs):
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		await self.close()

	def openable(self):
		raise NotImplementedError

	async def chmod(self, path: str, mode: int, **kwargs):
		raise NotImplemented

	async def chown(self, path: str, uid: int, gid: int, **kwargs):
		raise NotImplemented

	async def close(self, **kwargs):
		raise NotImplementedError

	async def open(self, filename: str, mode: str = 'r', **kwargs):
		raise NotImplemented

	async def get(self, remotepath: str, localpath: str, **kwargs):
		raise NotImplementedError

	async def put(self, localpath: str, remotepath: str, **kwargs):
		raise NotImplementedError

	async def listdir(self, path='.', **kwargs):
		raise NotImplemented

	async def mkdir(self, path, mode=511, **kwargs):
		raise NotImplemented

	async def remove(self, path: str, **kwargs):
		raise NotImplemented

	async def rename(self, oldpath: str, newpath: str, **kwargs):
		raise NotImplemented

	async def rmdir(self, path: str, **kwargs):
		raise NotImplemented

	async def stat(self, path: str, **kwargs):
		raise NotImplemented

	async def exists(self, path: str, **kwargs):
		raise NotImplementedError

	async def is_file(self, path: str, **kwargs):
		raise NotImplementedError

	async def is_dir(self, path: str, **kwargs):
		raise NotImplementedError

	async def is_link(self, path: str, **kwargs):
		raise NotImplementedError

	async def symlink(self, source: str, dest: str, **kwargs):
		raise NotImplemented

	async def touch(self, path: str, **kwargs):
		raise NotImplementedError



"""
This file contains the apps & apps settings and overrides the default ones that are defined in the core.
Documentation: http://pypla.net/

If you want to use other configuration methods like YAML or JSON files, take a look at http://pypla.net/ and head to the
configuration pages.
"""

# In the apps dictionary and array you configure the apps (or plugins) are loaded for specific pools (controllers).
# Be aware that the list will *ALWAYS* be prepended after the mandatory defaults are loaded in place.
# The mandatory defaults are specific per version, refer to the documentation:
# DOCUMENTATION: http://pypla.net/
APPS = {
	'default': [
		'pyplanet.apps.contrib.admin',
		'pyplanet.apps.contrib.jukebox',
		'pyplanet.apps.contrib.karma',
		'pyplanet.apps.contrib.local_records',
		'pyplanet.apps.contrib.dedimania',  # Will be disabled in Shootmania automatically.
		'pyplanet.apps.contrib.players',
		'pyplanet.apps.contrib.info',
		'pyplanet.apps.contrib.mx',
		'pyplanet.apps.contrib.transactions',
		'pyplanet.apps.contrib.sector_times',
		'pyplanet.apps.contrib.clock',
		'pyplanet.apps.contrib.funcmd',

		# You can optionally enable one of the following apps. Please look at this page for more information:
		# http://pypla.net/#app-docs

		# Live Ranking App. Useful when playing in Laps, Rounds and all sort of Trackmania game modes.
		'pyplanet.apps.contrib.live_rankings',

		# Best CP Widget on top of the screen for the Trackmania game.
		'pyplanet.apps.contrib.best_cps',

		# Use chat-based votes instead of the callvotes of the dedicated server with the voting app.
		'pyplanet.apps.contrib.voting',

		# Dynamic Points Limit is meant for Shootmania Royal.
		# 'pyplanet.apps.contrib.dynamic_points',

		# Waiting Queue. Enable on limited player servers to fairly queue players.
		# 'pyplanet.apps.contrib.queue',

		# Music Server App. Enable to queue your music together on the server.
		# 'pyplanet.apps.contrib.music_server',
	]
}



import os
import re
from setuptools import setup, find_packages


def long_description():
	try:
		return open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
	except IOError:
		return None


def read_version():
	with open(os.path.join(os.path.dirname(__file__), 'pyplanet', '__init__.py')) as handler:
		return re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", handler.read(), re.M).group(1)


def read_requirements(filename):
	with open(os.path.join(os.path.dirname(__file__), filename), 'r') as handler:
		return [line for line in handler.readlines() if not line.startswith('#') and not line.startswith('-') and not len(line) <= 1]


EXCLUDE_FROM_PACKAGES = [
	'pyplanet.bin',
	'pyplanet.conf.app_template',
	'pyplanet.conf.project_template',
	'docs*',
	'env*',
	'tests*',
	'apps*',
	'settings*',
]

PKG = 'pyplanet'
######
setup(
	name=PKG,
	version=read_version(),
	description='Maniaplanet Server Controller',
	long_description=long_description(),
	keywords='maniaplanet, controller, plugins, apps',
	license='GNU General Public License v3 (GPLv3)',
	packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
	package_data={
		'pyplanet/tests': ['pyplanet/tests/**.txt', 'pyplanet/tests/**.json', 'pyplanet/tests/**.xml', 'pyplanet/tests/**.j2']
	},
	install_requires=read_requirements('requirements.txt'),
	tests_require=read_requirements('requirements-dev.txt'),
	extras_require={},
	test_suite='tests',
	include_package_data=True,

	scripts=['pyplanet/bin/pyplanet'],
	entry_points={'console_scripts': [
		'pyplanet = pyplanet.core.management:execute_from_command_line',
	]},

	author='Tom Valk',
	author_email='tomvalk@lt-box.info',
	url='http://pypla.net/',

	classifiers=[
		'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
		'Development Status :: 4 - Beta',

		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3 :: Only',

		'Operating System :: OS Independent',

		'Topic :: Internet',
		'Topic :: Software Development',
		'Topic :: Games/Entertainment',
		'Topic :: Software Development :: Libraries :: Application Frameworks',
		'Topic :: Software Development :: Libraries :: Python Modules',

		'Intended Audience :: System Administrators',
		'Intended Audience :: Developers',

	],
	zip_safe=False,
)



"""
Mode contrib is managing mode settings and ui settings for the script mode.
"""
from .manager import ModeManager
from .signals import script_mode_changed

__all__ = [
	'ModeManager',
	'script_mode_changed'
]



import argparse
import logging.config
import threading

from pyplanet.conf import settings
from pyplanet.utils.log import initiate_logger
from pyplanet.god.pool import EnvironmentPool
from pyplanet import __version__ as version


class Management:
	def __init__(self, argv=None):
		self.parser = argparse.ArgumentParser()
		self.arguments = object()
		self.logger = logging.getLogger(__name__)
		self.add_arguments()

	def add_arguments(self):
		self.parser.add_argument('--settings')
		self.parser.add_argument('--max-restarts', type=int, default=0)
		# TODO.

	def execute(self):
		# Parse arguments.
		self.arguments = self.parser.parse_args()

		# Initiate the logger.
		threading.current_thread().setName('Main')
		initiate_logger()
		self.logger = logging.getLogger(__name__)

		# Initiate the settings by accessing one.
		self.logger.debug('Initiated configuration and environment... (debug on, means no error reporting and verbose output')
		if not settings.DEBUG:
			self.logger.info('Initiated configuration and environment...')
		self.logger.info('-------------------------------[  PyPlanet v{}  ]-------------------------------'.format(version))

		# Start god process (the current started process).
		pool = EnvironmentPool(settings.POOLS, max_restarts=self.arguments.max_restarts)
		pool.populate()

		# Starting all processes.
		pool.start()

		# Start the watchdog.
		try:
			pool.watchdog()
		except KeyboardInterrupt:
			pool.shutdown()
			exit(0)


def start(argv=None):
	"""
	Run the utility from CLI.
	:param argv: arguments.
	"""
	utility = Management(argv)
	utility.execute()



import threading
import multiprocessing

from colorlog import ColoredFormatter


def _run(name, queue, options):
	"""
	The actual process that runs the separate controller instance.

	:param name: name of the process
	:param queue: Queue of the binding parent.
	:param options: Custom Options
	:type name: str
	"""
	from pyplanet.core.instance import Controller
	from pyplanet.utils.log import initiate_logger, QueueHandler
	import logging

	# Tokio Asyncio (EXPERIMENTAL).
	if 'tokio' in options and options['tokio'] is True:
		import asyncio
		import tokio
		policy = tokio.TokioLoopPolicy()
		asyncio.set_event_loop_policy(policy)
		asyncio.set_event_loop(tokio.new_event_loop())
		logging.warning('Using experimental Tokio Asyncio Loop!')

	# Logging to queue.
	if multiprocessing.get_start_method() != 'fork':  # pragma: no cover
		initiate_logger()
		root_logger = logging.getLogger()
		formatter = ColoredFormatter(
			'%(log_color)s%(levelname)-8s%(reset)s %(yellow)s[%(threadName)s][%(name)s]%(reset)s %(blue)s%(message)s'
		)
		queue_handler = QueueHandler(queue)
		queue_handler.setFormatter(formatter)
		root_logger.addHandler(queue_handler)

	logging.getLogger(__name__).info('Starting pool process for \'{}\'...'.format(name))

	# Setting thread name to our process name.
	threading.main_thread().setName(name)

	# Initiate instance.
	instance = Controller.prepare(name).instance
	instance._queue = queue

	# Start and loop instance.
	instance.start()


class InstanceProcess:
	"""
	The InstanceProcess is the encapsulation around the real controller instance.

	.. warning::

		This code is still being executed at the main process!!

	"""

	def __init__(self, queue, environment_name='default', pool=None, options=None):
		"""
		Create an environment process of the controller itself.

		:param queue: Queue to hook on.
		:param environment_name: Name of environment.
		:param pool: Pool.
		:param options: Custom options.
		:type queue: multiprocessing.Queue
		:type environment_name: str
		:type pool: multiprocessing.Pool
		:type options: dict
		"""
		self.queue = queue
		self.name = environment_name
		self.options = options or dict()

		self.max_restarts = 1
		self.restarts = 0

		self.process = multiprocessing.Process(target=_run, kwargs=dict(
			name=self.name,
			queue=self.queue,
			options=self.options,
		))

		self.__last_state = True

	@property
	def did_die(self):
		"""
		Boolean determinating if the process did die.
		"""
		if not self.is_alive() and self.__last_state:
			self.__last_state = False
			return True
		return False

	@property
	def exitcode(self):
		"""
		Exit code of process.

		:return: Exit code.
		"""
		return self.process.exitcode

	@property
	def will_restart(self):
		"""
		Boolean: Is the process able to restart (not reached max_restarts).
		"""
		return self.restarts < self.max_restarts

	def is_alive(self):
		"""
		Call process method is_alive()
		"""
		return self.process.is_alive()

	def start(self):
		"""
		Start the process.
		"""
		return self.process.start()

	def shutdown(self):
		"""
		Shutdown (terminate) process.
		"""
		try:
			return self.process.terminate()
		except:
			pass
		return None

	def graceful(self):
		"""
		Graceful shutdown the process.
		"""
		self.process.join(timeout=10)



"""
Maniaplanet Core Models. This models are used in several apps and should be considered as very stable.
"""
from peewee import *
from pyplanet.core.db import TimedModel
from pyplanet.apps.core.maniaplanet.models import Player, Map


class MapFolder(TimedModel):
	player = ForeignKeyField(Player, index=True)
	"""
	The player who created the folder.
	"""

	name = CharField(
		max_length=50,
		null=False,
	)
	"""
	Folder name
	"""

	public = BooleanField()
	"""
	Is folder public (only for admins)?
	"""

	@property
	def icon(self):
		if self.public:
			return '\uf0c0'
		return '\uf023'


class MapInFolder(TimedModel):
	map = ForeignKeyField(Map, index=True)
	"""
	Map belonging to a folder.
	"""

	folder = ForeignKeyField(MapFolder, index=True)
	"""
	Folder the map belongs to.
	"""




import datetime

from peewee import Model as PeeweeModel, ReverseRelationDescriptor
from peewee import DateTimeField
from peewee_async import Manager

from .database import Proxy


class ObjectManager(Manager):
	database = Proxy

	async def get_related(self, instance, related_name, single_backref=False):
		"""
		return related instance for foreign key relationship
		return query for backref or return related instance if single_backref is True
		"""
		model_cls = type(instance)
		related_field = getattr(model_cls, related_name)

		if isinstance(related_field, ReverseRelationDescriptor):
			return await self._get_backrefs(instance, related_name, single_backref)
		else:
			return await self._get_foreign_key_target(instance, related_name)

	async def _get_foreign_key_target(self, instance, field_name):
		foreign_key_value = getattr(instance, field_name + "_id")

		model_cls = type(instance)
		foreign_key_field = getattr(model_cls, field_name)
		target_cls = foreign_key_field.rel_model
		target_field = foreign_key_field.to_field

		return await self.get(target_cls, target_field == foreign_key_value)

	async def _get_backrefs(self, instance, related_name, single_backref=False):
		query = getattr(instance, related_name)
		instances = await self.execute(query)

		if single_backref:
			for instance in instances:
				return instance
			raise query.model_class.DoesNotExist
		else:
			return instances


class Model(PeeweeModel):
	objects = ObjectManager()

	@classmethod
	async def get(cls, *args, **kwargs):
		return await cls.objects.get(cls, *args, **kwargs)

	@classmethod
	async def execute(cls, query):
		return await cls.objects.execute(query)

	@classmethod
	async def get_or_create(cls, *args, **kwargs):
		return await cls.objects.get_or_create(cls, *args, **kwargs)

	@classmethod
	async def create(cls, **insert):
		inst = cls(**insert)
		await inst.save(force_insert=True)
		inst._prepare_instance()
		return inst

	@classmethod
	async def _insert(cls, **insert):
		return await cls.objects.create(cls, **insert)

	async def _update(self, **update):
		return await self.objects.update(self, only=update)

	async def get_related(self, related_name, single_backref=False):
		return await self.objects.get_related(self, related_name, single_backref)

	async def save(self, force_insert=False, only=None):
		field_dict = dict(self._data)
		if self._meta.primary_key is not False:
			pk_field = self._meta.primary_key
			pk_value = self._get_pk_value()
		else:
			pk_field = pk_value = None
		if only:
			field_dict = self._prune_fields(field_dict, only)
		elif self._meta.only_save_dirty and not force_insert:
			field_dict = self._prune_fields(
				field_dict,
				self.dirty_fields)
			if not field_dict:
				self._dirty.clear()
				return False

		self._populate_unsaved_relations(field_dict)
		if pk_value is not None and not force_insert:
			if self._meta.composite_key:
				for pk_part_name in pk_field.field_names:
					field_dict.pop(pk_part_name, None)
			else:
				field_dict.pop(pk_field.name, None)
			return await self._update(__data=True, **field_dict)
		elif pk_field is None:
			await self._insert(**field_dict)
			rows = 1
		else:
			pk_from_cursor = await self._insert(**field_dict)
			if pk_from_cursor is not None:
				pk_value = pk_from_cursor
			self._set_pk_value(pk_value)
			rows = 1
		self._dirty.clear()
		return rows

	async def destroy(self, recursive=False, delete_nullable=False):
		return await self.objects.delete(self, recursive, delete_nullable)

	class Meta:
		database = Proxy


class TimedModel(Model):
	created_at = DateTimeField(default=datetime.datetime.now)
	updated_at = DateTimeField(default=datetime.datetime.now)

	async def save(self, *args, **kwargs):
		self.updated_at = datetime.datetime.now()
		return await super().save(*args, **kwargs)

	@classmethod
	async def create(cls, **insert):
		insert['updated_at'] = datetime.datetime.now()
		return await super().create(**insert)



"""
This file contains the basic settings and overrides the default ones that are defined in the core.
Documentation: http://pypla.net/

If you want to use other configuration methods like YAML or JSON files, take a look at http://pypla.net/ and head to the
configuration pages.
"""
import os

# Set the root_path to the root of our project.
ROOT_PATH = os.path.dirname(os.path.dirname(__file__))

# Set the temporary location for the project.
TMP_PATH = os.path.join(ROOT_PATH, 'tmp')

# Create temporary folder if not exists.
if not os.path.exists(TMP_PATH):
	os.mkdir(TMP_PATH)

# Enable debug mode to get verbose output, not report any errors and dynamically use the DEBUG in your code
# for extra verbosity of logging/output.
DEBUG = bool(os.environ.get('PYPLANET_DEBUG', False))

# Add your pools (the controller instances per dedicated here) or leave as it is to use a single instance only.
POOLS = [
	'default'
]

# Owners are logins of the server owners, the owners always get *ALL* the permissions in the system.
OWNERS = {
	'default': [
		'your-maniaplanet-login'
	]
}

# Allow self-upgrading the installation. Disable on shared servers with one installation (hosting environment)!
SELF_UPGRADE = True

# Databases configuration holds an dictionary with information of the database backend.
# Please refer to the documentation for all examples. http://pypla.net/
DATABASES = {
	'default': {
		'ENGINE': 'peewee_async.MySQLDatabase',
		'NAME': 'pyplanet',
		'OPTIONS': {
			'host': 'localhost',
			'user': 'root',
			'password': '',
			'charset': 'utf8mb4',
		}
	}
}

# Dedicated configuration holds the different dedicated servers that the instances will run on including the names of
# the instances.
DEDICATED = {
	'default': {
		'HOST': '127.0.0.1',
		'PORT': '5000',
		'USER': 'SuperAdmin',
		'PASSWORD': 'SuperAdmin',
	}
}

# Map configuration is a set of configuration options related to match settings etc.
# Matchsettings filename.
MAP_MATCHSETTINGS = {
	'default': 'maplist.txt',
}

# Blacklist file is managed by the dedicated server and will be loaded and writen to by PyPlanet once a
# player gets blacklisted. The default will be the filename Maniaplanet always uses and is generic.
BLACKLIST_FILE = {
	'default': 'blacklist.txt'
}

# The storage configuration contains the same instance mapping of the dedicated servers and is used
# to access the filesystem on the dedicated server location.
# Please refer to the documentation for more information. http://pypla.net/
STORAGE = {
	'default': {
		'DRIVER': 'pyplanet.core.storage.drivers.local.LocalDriver',
		'OPTIONS': {},
	}
}

# Define any cache backends that can be used by the core and the plugins to cache data.
# This is not yet implemented. As soon as it's implemented you can activate it with the documentation, available on
# http://pypla.net/
# CACHE = {
# 	'default': {
#
# 	}
# }

# Songs is a list of URLs to .ogg files of songs to be played by the music server.
SONGS = {
	'default': []
}



import asynctest
from playhouse.reflection import Introspector

from pyplanet.core import Controller
from pyplanet.core.db.models import migration


class TestConnection(asynctest.TestCase):
	async def test_connection(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		with instance.db.allow_sync():
			introspector = Introspector.from_database(instance.db.engine)
			db_name = introspector.get_database_name()
			assert db_name and len(db_name) > 0


class TestModelDiscovery(asynctest.TestCase):

	async def test_discovery(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		assert len(instance.db.registry.models.keys()) > 0
		assert len(instance.db.registry.app_models.keys()) > 0


class TestModelTableCreation(asynctest.TestCase):

	async def test_creation(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		with instance.db.allow_sync():
			introspector = Introspector.from_database(instance.db.engine)
			metadata = introspector.introspect()
			assert len(metadata.model_names) > 0


class TestMigrations(asynctest.TestCase):

	async def test_db_migration(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		with instance.db.allow_sync():
			migration.Migration.create_table(True)
			await instance.db.migrator.migrate()

		# Reset to simulate second startup
		instance.db.migrator.pass_migrations = set()
		with instance.db.allow_sync():
			await instance.db.migrator.migrate()
		assert len(instance.db.migrator.pass_migrations) == 0



import async_generator
import asyncio_extras
import os
import shutil
import aiofiles

from pyplanet.core.storage import StorageDriver


class LocalDriver(StorageDriver):
	"""
	Local storage driver is using the Python build-in file access utilities for accessing a local storage-like system.
	
	:option BASE_PATH: Override the maniaplanet given base path.
	"""

	def __init__(self, instance, config: dict = None):
		super().__init__(instance, config)

		self.override_base_path = config['BASE_PATH'] if 'BASE_PATH' in config else None

	def absolute(self, path):
		if self.override_base_path:
			return os.path.join(self.override_base_path, path)
		return os.path.join(self.base_dir or '', path)

	async def chmod(self, path: str, mode: int, **kwargs):
		os.chmod(self.absolute(path), mode, **kwargs)

	async def chown(self, path: str, uid: int, gid: int, **kwargs):
		os.chown(self.absolute(path), uid, gid, **kwargs)

	async def close(self, **kwargs):
		pass

	@asyncio_extras.async_contextmanager
	async def open(self, filename: str, mode: str = 'r', **kwargs):
		fh = await aiofiles.open(self.absolute(filename), mode, **kwargs)
		await async_generator.yield_(fh)

	async def get(self, remotepath: str, localpath: str, **kwargs):
		return shutil.copy(src=self.absolute(remotepath), dst=localpath)

	async def put(self, localpath: str, remotepath: str, **kwargs):
		return shutil.copy(src=localpath, dst=self.absolute(remotepath))

	async def listdir(self, path='.', **kwargs):
		return os.listdir(self.absolute(path))

	async def mkdir(self, path, mode=511, **kwargs):
		os.mkdir(self.absolute(path), mode)

	async def remove(self, path: str, **kwargs):
		os.unlink(self.absolute(path), **kwargs)

	async def rename(self, oldpath: str, newpath: str, **kwargs):
		os.rename(self.absolute(oldpath), self.absolute(newpath))

	async def rmdir(self, path: str, **kwargs):
		shutil.rmtree(self.absolute(path), **kwargs)

	async def stat(self, path: str, **kwargs):
		return os.stat(self.absolute(path), **kwargs)

	async def exists(self, path: str, **kwargs):
		return os.path.exists(self.absolute(path))

	async def is_file(self, path: str, **kwargs):
		return os.path.isfile(self.absolute(path))

	async def is_dir(self, path: str, **kwargs):
		return os.path.isdir(self.absolute(path))

	async def is_link(self, path: str, **kwargs):
		return os.path.islink(self.absolute(path))

	async def symlink(self, source: str, dest: str, **kwargs):
		os.symlink(self.absolute(source), self.absolute(dest), **kwargs)

	async def touch(self, path: str, **kwargs):
		async with self.open(path, 'w+') as fh:
			await fh.write('')

	def openable(self):
		return True



import asyncio
from xmlrpc.client import Fault

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.jukebox.views import MapListView, JukeboxListView, FolderListView, AddToFolderView
from pyplanet.apps.contrib.jukebox.folders import FolderManager
from pyplanet.contrib.command import Command

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.setting import Setting
from pyplanet.views.generics.alert import show_alert


class Jukebox(AppConfig):
	name = 'pyplanet.apps.contrib.jukebox'
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.lock = asyncio.Lock()
		self.jukebox = []
		self.folder_manager = FolderManager(self)

		# Settings.
		self.setting_newest_days_range = Setting(
			'newest_days_range', 'Number of days for a map to be new', Setting.CAT_BEHAVIOUR, type=int,
			description='The number of days to a map should be visible in the newest mapfolder (defaults to 14 days).',
			default=14,
			change_target=self.folder_manager.update_folders
		)

		self.setting_allow_juking = Setting(
			'allow_juking', 'Allow juking of maps by players', Setting.CAT_BEHAVIOUR, type=bool,
			description='Allow juking maps by non-admin players.',
			default=True,
		)

	async def on_start(self):
		# Register permissions + commands.
		await self.instance.permission_manager.register('clear', 'Clear the jukebox', app=self, min_level=1)
		await self.instance.command_manager.register(
			Command(command='clearjukebox', aliases=['cjb'], target=self.clear_jukebox, perms='jukebox:clear', admin=True,
					description='Clears the current maps from the jukebox.'),
			Command(command='list', target=self.show_map_list,
					description='Displays the maps currently in the jukebox.').add_param(name='search', required=False),
			Command(command='jukebox', target=self.chat_command,
					description='Provides access to the jukebox commands.').add_param(name='option', required=False),
			Command(command='mapfolders', aliases=['mf'], target=self.show_map_folders,
					description='Shows the available (personal) map folders.'),
		)

		# Register settings.
		await self.context.setting.register(self.setting_newest_days_range, self.setting_allow_juking)

		# Register callback.
		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)

		# Add button to the map list for adding to a folder.
		MapListView.add_action(self.add_to_folder, 'Add to folder', '&#xf07b;', order=-50)

		# Fetch all folders.
		await self.folder_manager.on_start()

	def insert_map(self, player, map):
		self.jukebox.insert(0, {'player': player, 'map': map})

	def append_map(self, player, map):
		self.jukebox.append({'player': player, 'map': map})

	def empty_jukebox(self):
		self.jukebox.clear()
		self.lock.release()

	async def show_map_list(self, player, data, **kwargs):
		view = MapListView(self, player)
		if data.search is not None:
			view.search_text = data.search
		await view.display()

	async def show_map_folders(self, player, data, **kwargs):
		await self.folder_manager.display_folder_list(player)

	async def chat_command(self, player, data, **kwargs):
		if data.option is None:
			await self.display_chat_commands(player)
		else:
			async with self.lock:
				if data.option == 'list' or data.option == 'display':
					if len(self.jukebox) > 0:
						view = JukeboxListView(self)
						await view.display(player=player.login)
					else:
						message = '$i$f00There are currently no maps in the jukebox!'
						await self.instance.chat(message, player)

				elif data.option == 'drop':
					first_player = next((item for item in reversed(self.jukebox) if item['player'].login == player.login), None)
					if first_player is not None:
						self.jukebox.remove(first_player)
						message = '$fff{}$z$s$fa0 dropped $fff{}$z$s$fa0 from the jukebox.'.format(first_player['player'].nickname, first_player['map'].name)
						await self.instance.chat(message)
					else:
						message = '$i$f00You currently don\'t have a map in the jukebox.'
						await self.instance.chat(message, player)

				elif data.option == 'clear':
					if player.level == 0:
						message = '$i$f00You\'re not allowed to do this!'
						await self.instance.chat(message, player)
					else:
						if len(self.jukebox) > 0:
							self.jukebox.clear()
							message = '$ff0Admin $fff{}$z$s$ff0 has cleared the jukebox.'.format(player.nickname)
							await self.instance.chat(message)
						else:
							message = '$i$f00There are currently no maps in the jukebox.'
							await self.instance.chat(message, player)

				else:
					await self.display_chat_commands(player)

	async def display_chat_commands(self, player):
		message = '$ff0Available jukebox commands: $ffflist$ff0 | $fffdisplay$ff0 | $fffdrop$ff0'
		if player.level > 0:
			message += ' | $fffclear$ff0'
		message += '.'
		await self.instance.chat(message, player)

	async def clear_jukebox(self, player, data, **kwargs):
		async with self.lock:
			if len(self.jukebox) > 0:
				self.jukebox.clear()
				message = '$ff0Admin $fff{}$z$s$ff0 has cleared the jukebox.'.format(player.nickname)
				await self.instance.chat(message)
			else:
				message = '$i$f00There are currently no maps in the jukebox.'
				await self.instance.chat(message, player)

	async def add_to_jukebox(self, player, map):
		async with self.lock:
			if player.level == 0 and not await self.setting_allow_juking.get_value():
				message = '$i$f00Juking maps has been disabled for players!'
				await self.instance.chat(message, player)
				return

			if player.level == 0 and any(item['player'].login == player.login for item in self.jukebox):
				message = '$i$f00You already have a map in the jukebox! Wait till it\'s been played before adding another.'
				await self.instance.chat(message, player)
				return

			if map.get_id() == self.instance.map_manager.current_map.get_id() and player.level == 0:
				message = '$i$f00You can\'t add the current map to the jukebox!'
				await self.instance.chat(message, player)
				return

			if not any(item['map'] == map for item in self.jukebox):
				self.jukebox.append({'player': player, 'map': map})
				message = '$fff{}$z$s$fa0 was added to the jukebox by $fff{}$z$s$fa0.'.format(map.name, player.nickname)
				await self.instance.chat(message)
			else:
				message = '$i$f00This map has already been added to the jukebox, pick another one.'
				await self.instance.chat(message, player)

	async def drop_from_jukebox(self, player, instance):
		async with self.lock:
			if player.level == 0 and instance['player_login'] != player.login:
				message = '$i$f00You can only drop your own jukeboxed maps!'
				await self.instance.chat(message, player)
			else:
				drop_map = next((item for item in self.jukebox if item['map'].name == instance['map_name']), None)
				if drop_map is not None:
					self.jukebox.remove(drop_map)
					message = '$fff{}$z$s$fa0 dropped $fff{}$z$s$fa0 from the jukebox.'.format(player.nickname, instance['map_name'])
					await self.instance.chat(message)

	async def podium_start(self, **kwargs):
		if len(self.jukebox) == 0:
			return
		next = self.jukebox.pop(0)
		message = '$fa0The next map will be $fff{}$z$s$fa0 as requested by $fff{}$z$s$fa0.'.format(next['map'].name, next['player'].nickname)

		# Try to set the map, if not successful it might be that the map is removed while juked!
		try:
			await asyncio.gather(
				self.instance.chat(message),
				self.instance.map_manager.set_next_map(next['map'])
			)
		except Fault as e:
			# It's removed from the server.
			if 'Map not in the selection' in e.faultString or 'Map unknown' in e.faultString:
				await self.instance.chat(
					'$fa0Setting the next map has been canceled because the map is not on the server anymore!'
				)

				# Retry the next map(s).
				await self.podium_start()
			else:
				raise

	async def add_to_folder(self, player, values, map_dictionary, view, **kwargs):
		view = AddToFolderView(self, player, map_dictionary['id'], self.folder_manager)
		await view.display()
		result = await view.wait_for_response()
		await view.destroy()

		if result:
			await show_alert(player, 'Map has been added to the folder!', 'sm')

	def add_maplist_action(self):
		pass

	def remove_maplist_action(self):
		pass



"""
Mode Settings Views.
"""
import asyncio
from xmlrpc.client import Fault

from pyplanet.contrib.setting.exceptions import SerializationException
from pyplanet.views import TemplateView
from pyplanet.views.generics import ManualListView


class ModeSettingMenuView(ManualListView):
	title = 'Mode Settings'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'ProfileAdvanced'

	def __init__(self, app, player):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.contrib.admin.Admin
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		super().__init__()
		self.manager = app.context.ui
		self.app = app
		self.player = player
		self.sort_field = {
			'name': 'Name',
			'index': 'name',
			'sorting': True,
			'searching': True,
			'width': 60,
			'type': 'label'
		}

		self.child = None

	async def get_data(self):
		settings = dict()
		mode_info = await self.app.instance.mode_manager.get_current_script_info()
		mode_settings = await self.app.instance.mode_manager.get_settings()
		if 'ParamDescs' in mode_info:
			for info in mode_info['ParamDescs']:
				real_type = str
				if info['Type'] == 'boolean':
					real_type = bool
				elif info['Type'] == 'int':
					real_type = int
				elif info['Type'] == 'double':
					real_type = float

				settings[info['Name']] = dict(
					default=info['Default'], type=real_type, name=info['Name'], description=info['Desc']
				)

		for name, value in mode_settings.items():
			if name not in settings:
				settings[name] = dict(default=None, type=None, description='-')
			settings[name]['name'] = name
			settings[name]['value'] = value
			if not settings[name]['type']:
				settings[name]['type'] = type(value)
		return list(settings.values())

	async def open_edit_setting(self, player, values, row, **kwargs):
		if self.child:
			return

		# Show edit view.
		self.child = ModeSettingEditView(self, self.player, row)
		await self.child.display()
		await self.child.wait_for_response()
		await self.child.destroy()
		await self.display()  # refresh.
		self.child = None

	async def display(self, **kwargs):
		kwargs['player'] = self.player
		return await super().display(**kwargs)

	def value_renderer(self, row, field, **kwargs):
		value_type = type(row[field['index']])
		if row[field['index']] is None:
			return '-'
		if value_type == str or value_type == int or value_type == float or value_type == bool:
			return str(row[field['index']])
		elif value_type == dict:
			return 'Dictionary, edit to show'
		elif value_type == set or value_type == list:
			return '{} values, edit to show'.format(len(row[field['index']]))
		return 'Unknown type {}'.format(str(value_type))

	async def get_fields(self):
		return [
			{
				'name': 'Name',
				'index': 'name',
				'sorting': True,
				'searching': True,
				'width': 60,
				'type': 'label'
			},
			{
				'name': 'Description',
				'index': 'description',
				'sorting': False,
				'searching': True,
				'width': 75,
				'type': 'label',
			},
			{
				'name': 'Value or contents',
				'index': 'value',
				'sorting': False,
				'searching': False,
				'width': 70,
				'type': 'label',
				'renderer': self.value_renderer,
				'action': self.open_edit_setting
			},
		]

	async def get_actions(self):
		return [
			{
				'name': 'Edit',
				'type': 'label',
				'text': '$fff &#xf040; Edit',
				'width': 11,
				'action': self.open_edit_setting,
				'safe': True
			},
		]


class ModeSettingEditView(TemplateView):
	"""
	Mode Setting Edit view will provide the user with a friendly edit window.
	"""
	template_name = 'admin/setting/edit.xml'

	def __init__(self, parent, player, setting):
		"""
		Initiate child edit view.

		:param parent: Parent view.
		:param player: Player instance.
		:param setting: Setting dictionary.
		:type parent: pyplanet.view.base.View
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type setting: dict
		"""
		super().__init__(parent.manager)
		self.parent = parent
		self.player = player
		self.setting = setting

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)
		self.subscribe('button_save', self.save)
		self.subscribe('button_cancel', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()
		context['title'] = 'Edit \'{}\''.format(self.setting['name'])
		context['icon_style'] = 'Icons128x128_1'
		context['icon_substyle'] = 'ProfileAdvanced'

		context['setting'] = self.setting
		context['setting_value'] = self.setting['value']
		context['types'] = dict(int=int, str=str, float=float, set=set, dict=dict, list=list, bool=bool)

		return context

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])
		self.response_future.set_result(None)
		self.response_future.done()

	async def wait_for_response(self):
		return await self.response_future

	async def save(self, player, action, values, *args, **kwargs):
		value = values['setting_value_field']

		try:
			current_value = self.setting['value']
			current_type = type(current_value)

			if isinstance(current_value, bool):
				lower_setting_value = str(value).lower()

				if lower_setting_value == 'true' or value == '1':
					value = True
				elif lower_setting_value == 'false' or value == '0':
					value = False
				else:
					raise ValueError
			value = current_type(value)

			await self.parent.app.instance.mode_manager.update_settings({
				self.setting['name']: value
			})
		except ValueError:
			message = '$i$f00Unable to cast "$fff{}$f00" to required type ($fff{}$f00) for "$fff{}$f00".'.format(
				value, self.setting['type'], self.setting['name']
			)
			await self.parent.appinstance.chat(message, player)
			return
		except Fault as exception:
			message = '$i$f00Unable to set "$fff{}$f00" to "$fff{}$f00": $fff{}$f00.'.format(
				self.setting['name'], value, exception
			)
			await self.parent.app.instance.chat(message, player)
			return
		finally:
			await asyncio.gather(
				self.parent.app.instance.chat(
					'$ff0Changed mode setting "$fff{}$ff0" to "$fff{}$ff0" (was: "$fff{}$ff0").'.format(
						self.setting['name'], value, self.setting['value']
					),
					player
				),
				self.hide([player.login])
			)
			self.response_future.set_result(self.setting)
			self.response_future.done()



"""
Setting Views. Based on the Contrib component Setting.
"""
import asyncio

from pyplanet.contrib.setting.exceptions import SerializationException
from pyplanet.views import TemplateView
from pyplanet.views.generics import ManualListView


class SettingMenuView(ManualListView):
	title = 'PyPlanet Settings'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'ProfileAdvanced'

	def __init__(self, app, player):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		super().__init__()
		self.manager = app.context.ui
		self.app = app
		self.player = player

		self.child = None

	async def get_data(self):
		return [
			dict(
				key=setting.key,
				name=setting.name,
				app_label=setting.app_label,
				category=setting.category,
				type=setting.type,
				type_name=setting.type_name,
				value=setting._value[1],
			)
			for setting in await self.app.instance.setting_manager.get_all(prefetch_values=True)
		]

	async def open_edit_setting(self, player, values, row, **kwargs):
		if self.child:
			return

		# Getting the setting itself.
		setting = await self.app.instance.setting_manager.get_setting(row['app_label'], row['key'], prefetch_values=True)

		# Show edit view.
		self.child = SettingEditView(self, self.player, setting)
		await self.child.display()
		await self.child.wait_for_response()
		await self.child.destroy()
		await self.display()  # refresh.
		self.child = None

	async def display(self, **kwargs):
		kwargs['player'] = self.player
		return await super().display(**kwargs)

	def value_renderer(self, row, field, **kwargs):
		if row[field['index']] is None:
			return '-'
		if row['type'] == str or row['type'] == int or row['type'] == float or row['type'] == bool:
			return str(row[field['index']])
		elif row['type'] == dict:
			return 'Dictionary, edit to show'
		elif row['type'] == set or row['type'] == list:
			return '{} values, edit to show'.format(len(row[field['index']]))
		return 'Unknown type {}'.format(row['type_name'])

	async def get_fields(self):
		return [
			{
				'name': 'Name',
				'index': 'name',
				'sorting': True,
				'searching': True,
				'width': 50,
				'type': 'label'
			},
			{
				'name': 'App',
				'index': 'app_label',
				'sorting': True,
				'searching': True,
				'width': 20,
				'type': 'label'
			},
			{
				'name': 'Category',
				'index': 'category',
				'sorting': True,
				'searching': True,
				'width': 20,
				'type': 'label'
			},
			{
				'name': 'Type',
				'index': 'type_name',
				'sorting': True,
				'searching': True,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Value or contents',
				'index': 'value',
				'sorting': False,
				'searching': False,
				'width': 100,
				'type': 'label',
				'renderer': self.value_renderer,
				'action': self.open_edit_setting
			},
		]

	async def get_actions(self):
		return [
			{
				'name': 'Edit',
				'type': 'label',
				'text': '$fff &#xf040; Edit',
				'width': 11,
				'action': self.open_edit_setting,
				'safe': True
			},
		]


class SettingEditView(TemplateView):
	"""
	Setting Edit view will provide the user with a friendly edit window.
	"""
	template_name = 'core.pyplanet/setting/edit.xml'

	def __init__(self, parent, player, setting):
		"""
		Initiate child edit view.

		:param parent: Parent view.
		:param player: Player instance.
		:param setting: Setting instance.
		:type parent: pyplanet.view.base.View
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type setting: pyplanet.contrib.setting.setting.Setting
		"""
		super().__init__(parent.manager)
		self.parent = parent
		self.player = player
		self.setting = setting

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)
		self.subscribe('button_save', self.save)
		self.subscribe('button_cancel', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()
		app = None
		try:
			app = self.parent.app.instance.apps.apps[self.setting.app_label]
		except:
			pass

		context['title'] = 'Edit \'{}\''.format(self.setting)
		context['icon_style'] = 'Icons128x128_1'
		context['icon_substyle'] = 'ProfileAdvanced'

		context['app'] = app
		context['setting'] = self.setting
		context['setting_value'] = await self.setting.get_value(refresh=True)
		context['types'] = dict(int=int, str=str, float=float, set=set, dict=dict, list=list, bool=bool)

		return context

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])
		self.response_future.set_result(None)
		self.response_future.done()

	async def wait_for_response(self):
		return await self.response_future

	async def save(self, player, action, values, *args, **kwargs):
		raw_value = values['setting_value_field']

		# Convert list.
		raw_list = None
		if self.setting.type == list or self.setting.type == set:
			raw_list = str(raw_value).splitlines()
			if self.setting.type == set:
				raw_list = set(raw_list)

		try:
			if self.setting.type == list or self.setting.type == set:
				await self.setting.set_value(raw_list if len(raw_list) > 0 else None)
			else:
				await self.setting.set_value(raw_value)
		except SerializationException as e:
			await self.parent.app.instance.chat(
				'$fa0Error with saving setting: {}'.format(str(e)),
				player
			)
		except Exception as e:
			await self.parent.app.instance.chat(
				'$fa0Error with saving setting: {}'.format(str(e)),
				player
			)
		finally:
			await asyncio.gather(
				self.parent.app.instance.chat(
					'$fa0Setting has been saved \'{}\''.format(self.setting.key),
					player
				),
				self.hide([player.login])
			)
			self.response_future.set_result(self.setting)
			self.response_future.done()



import datetime

from pyplanet.apps.contrib.karma.models import Karma
from pyplanet.apps.contrib.local_records.models import LocalRecord
from pyplanet.apps.core.maniaplanet.models import Player, Map
from pyplanet.apps.core.statistics.models import Score
from pyplanet.contrib.converter.base import BaseConverter


class Xaseco2Converter(BaseConverter):
	"""
	The XAseco2 Converter uses MySQL to convert the data to the new PyPlanet structure.
	
	Please take a look at :doc:`Migrating from other controllers </convert/index>` pages for information on how to use
	this.
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.player_cache = dict()
		self.map_cache = dict()

	async def migrate(self, _):
		print('Migrating players...')
		await self.migrate_players()

		print('Migrating maps...')
		await self.migrate_maps()

		print('Migrating records...')
		await self.migrate_local_records()

		print('Migrating karma...')
		await self.migrate_karma()

		print('Migrating times...')
		await self.migrate_times()

	async def migrate_players(self):
		with self.connection.cursor() as cursor:
			cursor.execute('SELECT * FROM players')
			for s_player in cursor.fetchall():
				# Check if we already have this player in our database. If we have, ignore and print message.
				try:
					player = await Player.get_by_login(s_player['Login'])
					self.player_cache[player.login] = player

					print('Player with login \'{}\' already exists, skipping..'.format(s_player['Login']))
					continue
				except:
					# Not found, create it:
					player = await Player.create(
						login=s_player['Login'], nickname=s_player['NickName']
					)
					self.player_cache[player.login] = player

	async def migrate_maps(self):
		with self.connection.cursor() as cursor:
			cursor.execute('SELECT * FROM maps')
			for s_map in cursor.fetchall():
				# Check if we already have this player in our database. If we have, ignore and print message.
				try:
					map_instance = await Map.get_by_uid(s_map['Uid'])
					self.map_cache[map_instance.uid] = map_instance

					print('Map with uid \'{}\' already exists, skipping..'.format(s_map['Uid']))
					continue
				except:
					# Not found, create it:
					# HACK: We don't know the file yet. Empty string to fill until pyplanet has started next time.
					map_instance = await Map.create(
						uid=s_map['Uid'], name=s_map['Name'], file='', author_login=s_map['Author'],
						environment=s_map['Environment']
					)
					self.map_cache[map_instance.uid] = map_instance

	async def migrate_local_records(self):
		if 'local_records' not in self.instance.apps.apps:
			print('Skipping local records. App not activated!')
			return

		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT records.*, maps.Uid, players.Login '
				'FROM records, maps, players '
				'WHERE records.MapId = maps.Id AND records.PlayerId = players.Id'
			)
			for s_record in cursor.fetchall():
				# Get map + player
				try:
					map = self.map_cache[s_record['Uid']] if s_record['Uid'] in self.map_cache else await Map.get(uid=s_record['Uid'])
					player = self.player_cache[s_record['Login']] if s_record['Login'] in self.player_cache else await Player.get(login=s_record['Login'])
				except:
					# Skip.
					print('Can\'t convert record, map or player not found. Skipping...')
					continue

				try:
					await LocalRecord.get(map=map, player=player)
					print('Record with uid \'{}\' and player \'{}\' already exists, skipping..'.format(map.uid, player.login))
				except:
					await LocalRecord.create(
						map=map, player=player, score=s_record['Score'], checkpoints=s_record['Checkpoints'],
						created_at=s_record['Date'], updated_at=datetime.datetime.now()
					)

	async def migrate_karma(self):
		if 'karma' not in self.instance.apps.apps:
			print('Skipping karma. App not activated!')
			return

		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT rs_karma.*, maps.Uid, players.Login '
				'FROM rs_karma, maps, players '
				'WHERE rs_karma.MapId = maps.Id AND rs_karma.PlayerId = players.Id'
			)
			for s_karma in cursor.fetchall():
				# Get map + player
				try:
					map = self.map_cache[s_karma['Uid']] if s_karma['Uid'] in self.map_cache else await Map.get(uid=s_karma['Uid'])
					player = self.player_cache[s_karma['Login']] if s_karma['Login'] in self.player_cache else await Player.get(login=s_karma['Login'])
				except:
					# Skip.
					print('Can\'t convert karma, map or player not found. Skipping...')
					continue

				if s_karma['Score'] == 0:
					continue

				try:
					await Karma.get(map=map, player=player)
					print('Karma with uid \'{}\' and player \'{}\' already exists, skipping..'.format(map.uid, player.login))
				except:
					await Karma.create(
						map=map, player=player, score=-1 if s_karma['Score'] < 0 else 1,
						updated_at=datetime.datetime.now()
					)

	async def migrate_times(self):
		with self.connection.cursor() as cursor:
			cursor.execute(
				'SELECT rs_times.*, maps.Uid, players.Login '
				'FROM rs_times, maps, players '
				'WHERE rs_times.MapId = maps.Id AND rs_times.PlayerId = players.Id'
			)
			for s_time in cursor.fetchall():
				# Get map + player
				try:
					map = self.map_cache[s_time['Uid']] if s_time['Uid'] in self.map_cache else await Map.get(uid=s_time['Uid'])
					player = self.player_cache[s_time['Login']] if s_time['Login'] in self.player_cache else await Player.get(login=s_time['Login'])
				except:
					# Skip.
					print('Can\'t convert time, map or player not found. Skipping...')
					continue

				if s_time['Score'] == 0:
					continue

				try:
					await Score.get(
						map=map, player=player, score=s_time['Score'], created_at=datetime.datetime.fromtimestamp(s_time['Date'])
					)
					print('Score with uid \'{}\', player \'{}\', score \'{}\' at \'{}\' already exists, skipping..'.format(
						map.uid, player.login, s_time['Score'], s_time['Date'],
					))
				except:
					await Score.create(
						map=map, player=player, score=s_time['Score'], checkpoints=s_time['Checkpoints'],
						created_at=datetime.datetime.fromtimestamp(s_time['Date']),
					)




class ImproperlyConfigured(Exception):
	"""The configuration is not given or is invalid."""
	pass


class AppRegistryNotReady(Exception):
	"""The registry was not yet ready to invoke"""
	pass


class InvalidAppModule(Exception):
	"""The given app string is invalid or the app itself is misconfigured!"""


class TransportException(Exception):
	"""The XML-RPC tunnel got a transport error."""


class SignalException(Exception):
	"""Signal receiver thrown an exception!"""


class SignalGlueStop(Exception):
	"""Throw this exception inside of your glue method to stop executing the signal."""



# Levels
DEBUG = 10
INFO = 20
WARNING = 30
ERROR = 40
CRITICAL = 50


class CheckMessage:

	def __init__(self, level, msg, hint=None, obj=None, id=None):
		assert isinstance(level, int), "The first argument should be level."
		self.level = level
		self.msg = msg
		self.hint = hint
		self.obj = obj
		self.id = id

	def __eq__(self, other):
		return (
			isinstance(other, self.__class__) and
			all(getattr(self, attr) == getattr(other, attr)
				for attr in ['level', 'msg', 'hint', 'obj', 'id'])
		)

	def __str__(self):
		from pyplanet.core.db import model

		if self.obj is None:
			obj = "?"
		elif isinstance(self.obj, model.Model):
			# We need to hardcode ModelBase and Field cases because its __str__
			# method doesn't return "applabel.modellabel" and cannot be changed.
			obj = self.obj._meta.label
		else:
			obj = str(self.obj)
		id = "(%s) " % self.id if self.id else ""
		hint = "\n\tHINT: %s" % self.hint if self.hint else ''
		return "%s: %s%s%s" % (obj, id, self.msg, hint)

	def __repr__(self):
		return "<%s: level=%r, msg=%r, hint=%r, obj=%r, id=%r>" % \
			   (self.__class__.__name__, self.level, self.msg, self.hint, self.obj, self.id)

	def is_serious(self, level=ERROR):
		return self.level >= level

	def is_silenced(self):
		try:
			from pyplanet.conf import settings
			return self.id in settings.SILENCED_SYSTEM_CHECKS
		except:
			return False


class Debug(CheckMessage):
	def __init__(self, *args, **kwargs):
		super().__init__(DEBUG, *args, **kwargs)


class Info(CheckMessage):
	def __init__(self, *args, **kwargs):
		super().__init__(INFO, *args, **kwargs)


class Warning(CheckMessage):
	def __init__(self, *args, **kwargs):
		super().__init__(WARNING, *args, **kwargs)


class Error(CheckMessage):
	def __init__(self, *args, **kwargs):
		super().__init__(ERROR, *args, **kwargs)


class Critical(CheckMessage):
	def __init__(self, *args, **kwargs):
		super().__init__(CRITICAL, *args, **kwargs)




from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.admin.map import MapAdmin
from pyplanet.apps.contrib.admin.player import PlayerAdmin
from pyplanet.apps.contrib.admin.server import ServerAdmin
from pyplanet.apps.contrib.admin.pyplanet import PyPlanetAdmin
from pyplanet.apps.contrib.admin.flow import FlowAdmin
from pyplanet.apps.contrib.admin.toolbar import ToolbarAdmin
from pyplanet.apps.contrib.admin.mapbrowser import MapBrowser


class Admin(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.toolbar = ToolbarAdmin(self)
		self.player = PlayerAdmin(self)
		self.map = MapAdmin(self)
		self.server = ServerAdmin(self)
		self.pyplanet = PyPlanetAdmin(self)
		self.flow = FlowAdmin(self)
		self.mapbrowser = MapBrowser(self)

	async def on_start(self):
		await self.toolbar.on_start()
		await self.player.on_start()
		await self.map.on_start()
		await self.server.on_start()
		await self.pyplanet.on_start()
		await self.flow.on_start()
		await self.mapbrowser.on_start()



"""
Player Admin methods and functions.
"""
import asyncio
import logging
import secrets
from random import random

from pyplanet.apps.contrib.admin.views.players import PlayerListView
from pyplanet.conf import settings
from pyplanet.contrib.command import Command
from pyplanet.contrib.player.exceptions import PlayerNotFound
from pyplanet.views.generics.alert import show_alert


class PlayerAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance
		self.claim_token = None

	async def on_start(self):
		await self.instance.permission_manager.register('ignore', 'Ignore a player', app=self.app, min_level=1)
		await self.instance.permission_manager.register('unignore', 'Unignore a player', app=self.app, min_level=1)
		await self.instance.permission_manager.register('kick', 'Kick a player', app=self.app, min_level=1)
		await self.instance.permission_manager.register('ban', 'Ban a player', app=self.app, min_level=2)
		await self.instance.permission_manager.register('unban', 'Unban a player', app=self.app, min_level=2)
		await self.instance.permission_manager.register('blacklist', 'Blacklist a player', app=self.app, min_level=3)
		await self.instance.permission_manager.register('unblacklist', 'Unblacklist a player', app=self.app, min_level=3)
		await self.instance.permission_manager.register('manage_admins', 'Manage admin', app=self.app, min_level=3)
		await self.instance.permission_manager.register('list_players', 'List players', app=self.app, min_level=1)
		await self.instance.permission_manager.register('force_spec', 'Force player into spectate', app=self.app, min_level=1)
		await self.instance.permission_manager.register('force_play', 'Force player into player slot', app=self.app, min_level=1)
		await self.instance.permission_manager.register('force_team', 'Force player into a team', app=self.app, min_level=1)
		await self.instance.permission_manager.register('switch_team', 'Force player into the other team', app=self.app, min_level=1)
		await self.instance.permission_manager.register('warn', 'Warn a player', app=self.app, min_level=1)
		await self.instance.permission_manager.register('write_blacklist', 'Write blacklist file', app=self.app, min_level=3)
		await self.instance.permission_manager.register('read_blacklist', 'Read blacklist file', app=self.app, min_level=3)
		await self.instance.permission_manager.register('addguest', 'Guestlist a player', app=self.app, min_level=3)
		await self.instance.permission_manager.register('write_guestlist', 'Read guestlist file', app=self.app, min_level=3)
		await self.instance.permission_manager.register('read_guestlist', 'Read guestlist file', app=self.app, min_level=3)
		await self.instance.permission_manager.register('removeguest', 'Remove a guest', app=self.app, min_level=3)

		await self.instance.command_manager.register(
			Command(command='players', target=self.player_list, perms='admin:list_players', admin=True,
					description='Displays the list of players currently online.'),
			Command(command='mute', aliases=['ignore'], target=self.ignore_player, perms='admin:ignore', admin=True,
					description='Mutes the provided player (sent chat messages will not be visible).').add_param(name='login', required=True),
			Command(command='unmute', aliases=['unignore'], target=self.unignore_player, perms='admin:unignore', admin=True,
					description='Unmutes the provided players (sent chat messages will be visible again).').add_param(name='login', required=True),
			Command(command='kick', target=self.kick_player, perms='admin:kick', admin=True,
					description='Kicks the provided player from the server.').add_param(name='login', required=True),
			Command(command='ban', target=self.ban_player, perms='admin:ban', admin=True,
					description='Bans the provided player from the server (until server restart).').add_param(name='login', required=True),
			Command(command='unban', target=self.unban_player, perms='admin:unban', admin=True,
					description='Unbans the provided player on the server (until server restart).').add_param(name='login', required=True),
			Command(command='blacklist', aliases=['black'], target=self.blacklist_player, perms='admin:blacklist', admin=True,
					description='Blacklists the provided player from the server (permanent ban).').add_param(name='login', required=True),
			Command(command='addguest', aliases=['addguest'], target=self.addguest_player, perms='admin:addguest', admin=True,
					description='Guests the provided player from the server (permanent guest).').add_param(name='login', required=True),
			Command(command='unblacklist', aliases=['unblack'], target=self.unblacklist_player, perms='admin:unblacklist', admin=True,
					description='Unblacklists the provided player on the server.').add_param(name='login', required=True),
			Command(command='removeguest', aliases=['removeguest'], target=self.removeguest_player, perms='admin:removeguest', admin=True,
					description='Remove the provided player on the server as a Guest.').add_param(name='login', required=True),
			Command(command='level', target=self.change_level, perms='admin:manage_admins', description='Changes admin level for the providedplayer.', admin=True)
				.add_param(name='login', required=True)
				.add_param(name='level', required=False, help='Level, 0 = player, 1 = operator, 2 = admin, 3 = master admin.', type=int, default=0),
			Command(command='forcespec', target=self.force_spec, perms='admin:force_spec', description='Forces player into spectator mode.', admin=True)
				.add_param(name='login', required=True),
			Command(command='forceplay', target=self.force_play, perms='admin:force_play', description='Forces player into player slot.', admin=True)
				.add_param(name='login', required=True),
			Command(command='forceteam', target=self.force_team, perms='admin:force_team', description='Forces player into a team.', admin=True)
				.add_param(name='login', required=True, type=str)
				.add_param(name='team', required=True, help='Team, blue/0 = left, red/1 = right'),
			Command(command='switchteam', target=self.switch_team, perms='admin:switch_team', description='Forces player into the other team.', admin=True)
				.add_param(name='login', required=True),
			Command(command='warn', aliases=['warning'], target=self.warn_player, perms='admin:warn', description='Warns the provided player.', admin=True)
				.add_param(name='login', required=True),
			Command(command='writeblacklist', aliases=['wbl'], target=self.write_blacklist, perms='admin:write_blacklist', description='Writes the blacklist file to disk.', admin=True)
				.add_param('file', required=False, type=str, help='Give custom blacklist file to save to.'),
			Command(command='writeguestlist', aliases=['wgl'], target=self.write_guestlist, perms='admin:write_guestlist', description='Writes the guestlist file to disk.', admin=True)
				.add_param('file', required=False, type=str, help='Give custom guest file to save to.'),
			Command(command='readblacklist', aliases=['rbl'], target=self.read_blacklist, perms='admin:read_blacklist', description='Reads the blacklist file from disk.', admin=True)
				.add_param('file', required=False, type=str, help='Give custom blacklist file to load from.'),
			Command(command='readguestlist', aliases=['rgl'], target=self.read_guestlist, perms='admin:read_guestlist', description='Reads the guestlist file from disk.', admin=True)
				.add_param('file', required=False, type=str, help='Give custom guestlist file to load from.'),
			Command(command='claim', target=self.claim_rights, description='Claim admin rights', admin=False)
				.add_param('token', required=True, type=str, help='Token to claim rights.'),
		)

		# Claim timer.
		self.claim_token = secrets.token_hex(8)
		asyncio.ensure_future(self.announce_claim_message())

	async def force_spec(self, player, data, **kwargs):
		try:
			dest_player = [p for p in self.instance.player_manager.online if p.login == data.login]
			if not len(dest_player) == 1:
				raise Exception()
			message = '$ff0Admin $fff{}$z$s$ff0 has forced $fff{}$z$s$ff0 into spectator.'.format(
				player.nickname, dest_player[0].nickname
			)
			await self.instance.gbx.multicall(
				self.instance.gbx('ForceSpectator', dest_player[0].login, 3),
				self.instance.chat(message)
			)
		except Exception:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)

	async def force_play(self, player, data, **kwargs):
		try:
			dest_player = [p for p in self.instance.player_manager.online if p.login == data.login]
			if not len(dest_player) == 1:
				raise Exception()
			message = '$ff0Admin $fff{}$z$s$ff0 has forced $fff{}$z$s$ff0 into player slot.'.format(
				player.nickname, dest_player[0].nickname
			)
			await self.instance.gbx('ForceSpectator', dest_player[0].login, 2)
			await self.instance.gbx.multicall(
				self.instance.gbx('ForceSpectator', dest_player[0].login, 0),
				self.instance.chat(message)
			)
		except Exception as e:
			if 'There are too many players' in str(e):
				message = '$i$f00Too many players!'
			else:
				message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)

	async def force_team(self, player, data, **kwargs):
		dest_player = [p for p in self.instance.player_manager.online if p.login == data.login]
		if not len(dest_player) == 1:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
			return

		new_team = None
		team_name = ''
		if data.team == 'blue' or data.team == '0':
			new_team = 0
			team_name = '$00fBLUE'
		elif data.team == 'red' or data.team == '1':
			new_team = 1
			team_name = '$f00RED'

		if new_team is None:
			message = '$i$f00Unknown team identifier ($fff{}$z$s$f00)!'.format(data.team)
			await self.instance.chat(message, player)
			return

		message = '$ff0Admin $fff{}$z$s$ff0 has forced $fff{}$z$s$ff0 into team $fff{}$ff0.'.format(
			player.nickname, dest_player[0].nickname, team_name
		)
		await self.instance.gbx.multicall(
			self.instance.gbx('ForcePlayerTeam', dest_player[0].login, new_team),
			self.instance.chat(message)
		)

	async def switch_team(self, player, data, **kwargs):
		dest_player = [p for p in self.instance.player_manager.online if p.login == data.login]
		if not len(dest_player) == 1:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
			return

		new_team = None
		team_name = ''
		if dest_player[0].flow.team_id == 1:
			new_team = 0
			team_name = '$00fBLUE'
		elif dest_player[0].flow.team_id == 0:
			new_team = 1
			team_name = '$f00RED'

		if new_team is None:
			message = '$i$f00Unable to switch team (are you in a team mode?)!'
			await self.instance.chat(message, player)
			return

		message = '$ff0Admin $fff{}$z$s$ff0 has forced $fff{}$z$s$ff0 into team $fff{}$ff0.'.format(
			player.nickname, dest_player[0].nickname, team_name
		)
		await self.instance.gbx.multicall(
			self.instance.gbx('ForcePlayerTeam', dest_player[0].login, new_team),
			self.instance.chat(message)
		)

	async def ignore_player(self, player, data, **kwargs):
		try:
			if data.login not in self.instance.player_manager.online_logins:
				raise PlayerNotFound()
			mute_player = await self.instance.player_manager.get_player(data.login)
			if mute_player.level >= player.level:
				raise PermissionError()
			message = '$ff0Admin $fff{}$z$s$ff0 has muted $fff{}$z$s$ff0.'.format(player.nickname, mute_player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('Ignore', data.login),
				self.instance.chat(message)
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
		except PermissionError:
			message = '$i$f00Can\'t perform this action on an admin at the same or higher level as you!'
			await self.instance.chat(message, player)

	async def unignore_player(self, player, data, **kwargs):
		try:
			if data.login not in self.instance.player_manager.online_logins:
				raise PlayerNotFound()

			unmute_player = await self.instance.player_manager.get_player(data.login)
			message = '$ff0Admin $fff{}$z$s$ff0 has un-muted $fff{}$z$s$ff0.'.format(player.nickname, unmute_player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('UnIgnore', data.login),
				self.instance.chat(message)
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)

	async def kick_player(self, player, data, **kwargs):
		try:
			if data.login not in self.instance.player_manager.online_logins:
				raise PlayerNotFound()
			kick_player = await self.instance.player_manager.get_player(data.login)
			if kick_player.level >= player.level:
				raise PermissionError()

			message = '$ff0Admin $fff{}$z$s$ff0 has kicked $fff{}$z$s$ff0.'.format(player.nickname, kick_player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('Kick', data.login),
				self.instance.chat(message)
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
		except PermissionError:
			message = '$i$f00Can\'t perform this action on an admin at the same or higher level as you!'
			await self.instance.chat(message, player)

	async def ban_player(self, player, data, **kwargs):
		try:
			ban_player = await self.instance.player_manager.get_player(data.login)
			if ban_player.level >= player.level:
				raise PermissionError()
			message = '$ff0Admin $fff{}$z$s$ff0 has banned $fff{}$z$s$ff0.'.format(player.nickname, ban_player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('Ban', data.login),
				self.instance.chat(message)
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
		except PermissionError:
			message = '$i$f00Can\'t perform this action on an admin at the same or higher level as you!'
			await self.instance.chat(message, player)

	async def unban_player(self, player, data, **kwargs):
		message = '$ff0Admin $fff{}$z$s$ff0 has un-banned $fff{}$z$s$ff0.'.format(player.nickname, data.login)
		await self.instance.gbx.multicall(
			self.instance.gbx('UnBan', data.login),
			self.instance.chat(message)
		)

	async def addguest_player(self, player, data, **kwargs):
		try:
			guest_player = await self.instance.player_manager.get_player(data.login)
			if guest_player.level >= player.level:
				raise PermissionError()
			message = '$ff0Admin $fff{}$z$s$ff0 has added to the Guestlist: $fff{}$z$s$ff0.'.format(player.nickname, guest_player.nickname)
			await self.instance.gbx.multicall(
				self.instance.gbx('AddGuest', data.login),
				self.instance.chat(message)
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)
		except PermissionError:
			message = '$i$f00Can\'t perform this action on an admin at the same or higher level as you!'
			await self.instance.chat(message, player)

	async def removeguest_player(self, player, data, **kwargs):
		message = '$ff0Admin $fff{}$z$s$ff0 has removed from the Guestlist: $fff{}$z$s$ff0.'.format(player.nickname, data.login)
		await self.instance.gbx.multicall(
			self.instance.gbx('RemoveGuest', data.login),
			self.instance.chat(message)
		)

	async def blacklist_player(self, player, data, **kwargs):
		try:
			blacklist_player = await self.instance.player_manager.get_player(data.login)
			if blacklist_player.level >= player.level:
				raise PermissionError()

			message = '$ff0Admin $fff{}$z$s$ff0 has blacklisted $fff{}$z$s$ff0.'.format(player.nickname, blacklist_player.nickname)


			try:
				await self.instance.gbx.multicall(
					self.instance.gbx('BlackList', data.login),
					self.instance.gbx('Kick', data.login),
				)
			except:
				return await self.instance.chat('$ff0Blacklisting failed!', player)

			await self.instance.chat(message)

			# Try to save to file.
			try:
				await self.instance.player_manager.save_blacklist()
			except:
				pass
		except PlayerNotFound:
			message = '$ff0Admin $fff{}$z$s$ff0 has blacklisted $fff{}$z$s$ff0.'.format(player.nickname, data.login)
			await self.instance.gbx.multicall(
				self.instance.gbx('BlackList', data.login),
				self.instance.chat(message)
			)
		except PermissionError:
			message = '$i$f00Can\'t perform this action on an admin at the same or higher level as you!'
			await self.instance.chat(message, player)

	async def unblacklist_player(self, player, data, **kwargs):
		message = '$ff0Admin $fff{}$z$s$ff0 has un-blacklisted $fff{}$z$s$ff0.'.format(player.nickname, data.login)
		await self.instance.gbx.multicall(
			self.instance.gbx('UnBlackList', data.login),
			self.instance.chat(message)
		)

		# Try to save to file.
		try:
			await self.instance.player_manager.save_blacklist()
		except:
			pass

	async def write_guestlist(self, player, data, **kwargs):
		setting = settings.GUESTLIST_FILE
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not setting and not data.file:
			message = '$ff0Default guestlist file setting not configured in your settings file!'
			await self.instance.chat(message, player)
			return

		if data.file:
			file_name = data.file
		else:
			file_name = setting.format(server_login=self.instance.game.server_player_login)

		message = '$ff0Guestlist has been saved to the file: {}'.format(file_name)
		try:
			await self.instance.player_manager.save_guestlist(filename=file_name)
			await self.instance.chat(message, player)
		except:
			await self.instance.chat('$ff0Guestlist saving failed to {}'.format(file_name), player)

	async def read_guestlist(self, player, data, **kwargs):
		setting = settings.GUESTLIST_FILE
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not setting and not data.file:
			message = '$ff0Default Guest list file setting not configured in your settings file!'
			await self.instance.chat(message, player)
			return

		if data.file:
			file_name = data.file
		else:
			file_name = setting.format(server_login=self.instance.game.server_player_login)

		message = '$ff0Guestlist has been loaded from the file: {}'.format(file_name)
		try:
			await self.instance.player_manager.load_guestlist(filename=file_name)
			await self.instance.chat(message, player)
		except:
			await self.instance.chat('$ff0Guestlist loading failed from {}'.format(file_name), player)


	async def change_level(self, player, data, **kwargs):
		try:
			target_player = await self.instance.player_manager.get_player(login=data.login)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player.login)
			return

		if data.level < 0 or data.level > 3:
			message = '$i$f00Level must be between 0 and 3! 0=player, 1=operator, 2=admin, 3=masteradmin.'
			await self.instance.chat(message, player)
			return

		old_level_name = target_player.get_level_string()
		target_player.level = data.level
		new_level_name = target_player.get_level_string()
		await target_player.save()

		if data.level > 0:
			message = '$ff0Admin $fff{}$z$s$ff0 has added $fff{}$z$s$ff0 as an {}.'.format(
				player.nickname, target_player.nickname, new_level_name
			)
		else:
			message = '$ff0Admin $fff{}$z$s$ff0 has removed $fff{}$z$s$ff0 as an {}.'.format(
				player.nickname, target_player.nickname, old_level_name
			)
		await self.instance.chat(message)

	async def warn_player(self, player, data, **kwargs):
		try:
			warn_player = await self.instance.player_manager.get_player(login=data.login, lock=False)

			await asyncio.gather(
				show_alert(
					warn_player,
					'You have just been warned! Ask the present admin for further information and / or potential consequences.',
					size='sm', buttons=None
				),
				self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has warned $fff{}$z$s$ff0.'.format(player.nickname, warn_player.nickname))
			)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player.login)
			return

	async def write_blacklist(self, player, data, **kwargs):
		setting = settings.BLACKLIST_FILE
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not setting and not data.file:
			message = '$ff0Default blacklist file setting not configured in your settings file!'
			await self.instance.chat(message, player)
			return

		if data.file:
			file_name = data.file
		else:
			file_name = setting.format(server_login=self.instance.game.server_player_login)

		message = '$ff0Blacklist has been saved to the file: {}'.format(file_name)
		try:
			await self.instance.player_manager.save_blacklist(filename=file_name)
			await self.instance.chat(message, player)
		except:
			await self.instance.chat('$ff0Blacklist saving failed to {}'.format(file_name), player)

	async def read_blacklist(self, player, data, **kwargs):
		setting = settings.BLACKLIST_FILE
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not setting and not data.file:
			message = '$ff0Default blacklist file setting not configured in your settings file!'
			await self.instance.chat(message, player)
			return

		if data.file:
			file_name = data.file
		else:
			file_name = setting.format(server_login=self.instance.game.server_player_login)

		message = '$ff0Blacklist has been loaded from the file: {}'.format(file_name)
		try:
			await self.instance.player_manager.load_blacklist(filename=file_name)
			await self.instance.chat(message, player)
		except:
			await self.instance.chat('$ff0Blacklist loading failed from {}'.format(file_name), player)

	async def player_list(self, player, data, **kwargs):
		view = PlayerListView(self.app, player)
		await view.display()

	async def claim_rights(self, player, data, **kwargs):
		if data.token != self.claim_token:
			await self.instance.chat('$ff0Claiming rights failed. Failure has been logged!', player)
			logging.getLogger(__name__).error('Security: User {} ({}) tried to claim rights'.format(player.nickname, player.login))
			return

		player.level = 3
		await player.save()
		await self.instance.chat('$fff{}$z$s$ff0 has claimed admin rights.'.format(player.nickname))

	async def announce_claim_message(self):
		await asyncio.sleep(4)
		logging.getLogger(__name__).info(
			'Welcome to PyPlanet, to claim admin rights, copy and paste this in the chat: /claim {}'.format(self.claim_token)
		)



import asyncio
import math

from playhouse.shortcuts import model_to_dict

from pyplanet.apps.contrib.jukebox.models import MapFolder, MapInFolder
from pyplanet.apps.core.maniaplanet.models import Map
from pyplanet.views import TemplateView
from pyplanet.views.generics.alert import show_alert, ask_confirmation, ask_input
from pyplanet.views.generics.list import ManualListView

from pyplanet.utils import times


class JukeboxListView(ManualListView):
	app = None

	title = 'Currently in the Jukebox'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	data = []

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'index',
				'sorting': True,
				'searching': False,
				'width': 10,
				'type': 'label'
			},
			{
				'name': 'Map',
				'index': 'map_name',
				'sorting': True,
				'searching': True,
				'width': 100,
				'type': 'label',
				'action': self.action_drop
			},
			{
				'name': 'Requested by',
				'index': 'player_nickname',
				'sorting': True,
				'searching': False,
				'width': 50
			},
		]

	async def action_drop(self, player, values, instance, **kwargs):
		await self.app.drop_from_jukebox(player, instance)
		await self.refresh(player=player)

	async def get_data(self):
		index = 1
		items = []
		for item in self.app.jukebox:
			items.append({'index': index, 'map_name': item['map'].name, 'player_nickname': item['player'].nickname,
						  'player_login': item['player'].login})
			index += 1

		return items


class MapListView(ManualListView):
	model = Map
	title = 'Maps on this server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	custom_actions = list()

	supports_advanced = True

	def __init__(self, app, player):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.player = player
		self.advanced = False
		self.cache = list()
		self.cache_advanced = False

	async def get_data(self):
		if self.cache and self.advanced == self.cache_advanced:
			return self.cache

		data = list()

		local_app_installed = 'local_records' in self.app.instance.apps.apps
		karma_app_installed = 'karma' in self.app.instance.apps.apps

		for m in self.app.instance.map_manager.maps:
			map_dict = model_to_dict(m)
			map_dict['local_record_rank'] = None
			map_dict['local_record_score'] = None
			map_dict['local_record_diff'] = None
			map_dict['local_record_diff_direction'] = None
			map_dict['karma'] = None

			# Skip if in performance mode or advanced is not enabled.
			if self.app.instance.performance_mode or not self.advanced:
				data.append(map_dict)
				continue

			if local_app_installed:
				# Get personal local record of the user.
				map_locals = await self.app.instance.apps.apps['local_records'].get_map_record(m)
				rank, record = await self.app.instance.apps.apps['local_records'].get_player_record_and_rank_for_map(m, self.player)

				if isinstance(rank, int) and rank >= 1:
					map_dict['local_record_rank'] = int(rank)
				if record:
					map_dict['local_record_score'] = record.score
				if map_locals['first_record'] and record:
					map_dict['local_record_diff'] = record.score - map_locals['first_record'].score

			# TODO: Convert to new relation styles.
			if karma_app_installed and hasattr(m, 'karma'):
				map_dict['karma'] = m.karma['map_karma']

			data.append(map_dict)

		self.cache = data
		self.cache_advanced = self.advanced
		return self.cache

	async def get_fields(self):
		fields = [
			{
				'name': 'Name',
				'index': 'name',
				'sorting': True,
				'searching': True,
				'search_strip_styles': True,
				'width': 90,
				'type': 'label',
				'action': self.action_jukebox
			},
			{
				'name': 'Author',
				'index': 'author_login',
				'sorting': True,
				'searching': True,
				'search_strip_styles': True,
				'renderer': lambda row, field:
				row['author_login'],
				# TODO: Activate after resolving #83.
				# row['author_nickname']
				# if 'author_nickname' in row and row['author_nickname'] and len(row['author_nickname'])
				# else row['author_login'],
				'width': 45,
			},
		]

		def render_optional_time(row, field):
			value = row[field['index']]
			if value is None:
				return ''
			if isinstance(value, (int, float)) and not math.isnan(value):
				return times.format_time(int(value))
			return 'None'

		def render_rank(row, field):
			value = row[field['index']]
			if value is None:
				return ''
			if isinstance(value, (int, float)) and not math.isnan(value):
				return int(value)
			return 'None'

		def render_karma(row, field):
			value = row[field['index']]
			if value is None:
				return ''
			if isinstance(value, (int, float)) and not math.isnan(value):
				prefix = ''
				if value > 0.0:
					prefix = '$6CF'
				elif value < 0.0:
					prefix = '$F66'
				return '{}{}'.format(prefix, float(value))

		if self.advanced and not self.app.instance.performance_mode:
			if 'karma' in self.app.instance.apps.apps:
				fields.append({
					'name': 'Karma',
					'index': 'karma',
					'sorting': True,
					'searching': False,
					'width': 15,
					'renderer': render_karma
				})

			if 'local_records' in self.app.instance.apps.apps:
				fields.append({
					'name': 'Rank',
					'index': 'local_record_rank',
					'sorting': True,
					'searching': False,
					'width': 15,
					'renderer': render_rank
				})
				fields.append({
					'name': 'Local',
					'index': 'local_record_score',
					'sorting': True,
					'searching': False,
					'width': 15,
					'renderer': render_optional_time
				})
				fields.append({
					'name': 'Diff #1',
					'index': 'local_record_diff',
					'sorting': True,
					'searching': False,
					'width': 15,
					'renderer': render_optional_time
				})

		return fields

	async def display(self, player=None):
		return await super().display(player or self.player)

	async def get_actions(self):
		return self.custom_actions

	async def get_buttons(self):
		buttons = [
			{
				'title': 'Folders',
				'width': 20,
				'action': self.action_folders
			}
		]

		if self.supports_advanced:
			if self.advanced:
				buttons.append({
					'title': 'Simple list',
					'width': 30,
					'action': self.action_advanced
				})
			else:
				buttons.append({
					'title': 'Advanced list',
					'width': 30,
					'action': self.action_advanced
				})

		return buttons

	async def action_jukebox(self, player, values, map_info, **kwargs):
		await self.app.add_to_jukebox(player, await self.app.instance.map_manager.get_map(map_info['uid']))

	async def action_folders(self, player, values, **kwargs):
		await self.app.folder_manager.display_folder_list(player)

	async def action_advanced(self, player, values, **kwargs):
		if len(self.app.instance.map_manager.maps) > 500:
			if self.player.level == 0:
				await self.app.instance.chat(
					'This server contains 500+ maps. Advanced map list only activated for admins!', self.player
				)
				return

		if self.advanced:
			self.advanced = False
		else:
			self.advanced = True
			await self.app.instance.chat(
				'Please wait... Loading advanced map list, this can take some time on large servers', self.player
			)
		await self.refresh(player=self.player)

	@classmethod
	def add_action(cls, target, name, text, text_size='1.2', require_confirm=False, order=0):
		cls.custom_actions.append(dict(
			name=name,
			action=target,
			text=text,
			textsize=text_size,
			safe=True,
			type='label',
			order=order,
			require_confirm=require_confirm,
		))

		cls.custom_actions = sorted(cls.custom_actions, key=lambda k: k['order'])

	@classmethod
	def remove_action(cls, target):
		for idx, custom in enumerate(cls.custom_actions):
			if custom['action'] == target:
				del cls.custom_actions[idx]

	async def destroy(self):
		await super().destroy()
		self.cache = list()

	def destroy_sync(self):
		super().destroy_sync()
		self.cache = list()


class FolderMapListView(MapListView):
	supports_advanced = False

	def __init__(self, folder_manager, folder_code, player):
		"""
		Folder Map list

		:param folder_manager: Folder manager
		:param folder_code: Folder code that can be parsed.
		:type folder_manager: pyplanet.apps.contrib.jukebox.folders.FolderManager
		"""
		super().__init__(folder_manager.app, player)

		self.folder_manager = folder_manager
		self.folder_code = folder_code
		self.player = player

		self.map_list = list()
		self.folder_info = None
		self.folder_instance = None

	async def get_fields(self):
		fields = await super().get_fields()

		for field in self.fields:
			fields.append(field)

		return fields

	async def get_data(self):
		if self.cache and self.advanced == self.cache_advanced:
			return self.cache

		self.fields, self.map_list, self.folder_info, self.folder_instance = \
			await self.folder_manager.get_folder_code_contents(self.folder_code)

		self.title = 'Folder: ' + self.folder_info['name']

		karma = any(f['index'] == "karma" for f in self.fields)
		length = any(f['index'] == "local_record" for f in self.fields)

		items = []
		for item in self.map_list:
			dict_item = model_to_dict(item)
			if length:
				dict_item['local_record'] = times.format_time((item.local['first_record'].score if hasattr(item, 'local') and item.local['first_record'] else 0))
			if karma and 'karma' in self.app.instance.apps.apps:
				dict_item['karma'] = (await self.app.instance.apps.apps['karma'].get_map_karma(item))['map_karma']
			items.append(dict_item)

		self.cache = items
		return self.cache

	async def remove_from_folder(self, player, values, map_dictionary, view, **kwargs):
		# Check permission on folder.
		if (self.folder_instance.public and player.level < player.LEVEL_ADMIN)\
			or (not self.folder_instance.public and self.folder_instance.player_id != player.id):
			await show_alert(player, 'You are not allowed to remove the map from the folder!', size='sm')
			return

		# Ask for confirmation.
		cancel = bool(await ask_confirmation(player, 'Are you sure you want to remove the map \'{}\'$z$s from the folder?'.format(
			map_dictionary['name']
		), size='sm'))
		if cancel is True:
			return

		# Remove from folder.
		await self.folder_manager.remove_map_from_folder(self.folder_instance.id, map_dictionary['id'])

		# Remove from the cache.
		for item in self.cache:
			if item['id'] == map_dictionary['id']:
				self.cache.remove(item)

		# Refresh list.
		await self.refresh(player)

	async def get_buttons(self):
		buttons = await super().get_buttons()

		if (self.folder_code['type'] == 'public' and self.player.level >= self.player.LEVEL_ADMIN) \
			or (self.folder_code['type'] == 'public' and self.folder_code['owner_login'] == self.player.login) \
			or (self.folder_code['type'] == 'private' and self.folder_code['owner_login'] == self.player.login):
			buttons.append({
				'title': 'Add current map',
				'width': 38,
				'action': self.action_add_current
			})

			buttons.append({
				'title': 'Rename folder',
				'width': 32,
				'action': self.action_rename
			})

		return buttons

	async def get_actions(self):
		super_actions = (await super().get_actions()).copy()

		# If this folder is not a dynamic one, add remove button.
		if self.folder_instance:
			super_actions.append(dict(
				name='Remove from folder',
				action=self.remove_from_folder,
				text='&#xf056;',
				textsize='1.2',
				safe=True,
				type='label',
				order=49,
				require_confirm=False,
			))

		return sorted(super_actions, key=lambda k: k['order'])

	async def action_add_current(self, player, values, **kwargs):
		if self.app.instance.map_manager.current_map.id in [m.id for m in self.map_list]:
			await show_alert(player, 'The current map is already in this folder!', 'sm')
			return

		# Add map to folder.
		map_in_folder = MapInFolder(
			map_id=self.app.instance.map_manager.current_map.id,
			folder=self.folder_instance
		)
		await map_in_folder.save()

		await show_alert(player, 'Map has been added to the folder!', 'sm')
		await self.refresh(player)

	async def action_rename(self, player, values, **kwargs):
		new_name = await ask_input(
			player, 'Please enter the new name of the folder', size='sm', default=self.folder_instance.name,
		)

		self.folder_instance.name = new_name
		await self.folder_instance.save()

		# Change the cached name so it shows when we refresh.
		if self.folder_info and 'name' in self.folder_info:
			self.folder_info['name'] = new_name

		await self.display(player)


class FolderListView(ManualListView):
	title = 'Maplist folders'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	def __init__(self, folder_manager, player):
		super().__init__()

		self.folder_manager = folder_manager
		self.player = player
		self.app = folder_manager.app
		self.manager = folder_manager.app.context.ui

		self.child = None

	@staticmethod
	def render_folder_name(row, field):
		icon = ''
		if row['type'] == 'auto':
			icon = '\uf013'
		elif row['type'] == 'public':
			icon = '\uf0c0'
		elif row['type'] == 'private':
			icon = '\uf023'

		return '{} {}'.format(icon, row[field['index']])

	async def get_fields(self):
		return [
			{
				'name': 'Folder',
				'index': 'name',
				'sorting': False,
				'searching': True,
				'width': 130,
				'renderer': self.render_folder_name,
				'type': 'label',
				'action': self.action_show
			},
			{
				'name': 'Type',
				'index': 'type',
				'sorting': True,
				'searching': False,
				'width': 30,
				'renderer': lambda row, field:
					row[field['index']].capitalize(),
				'type': 'label'
			},
			{
				'name': 'Owner',
				'index': 'owner',
				'sorting': False,
				'searching': False,
				'width': 80,
			},
		]

	async def get_buttons(self):
		return [
			{
				'title': 'Create folder',
				'width': 28,
				'action': self.create_folder
			}
		]

	async def get_actions(self):
		return [
			dict(
				name='Delete folder',
				action=self.delete_folder,
				text='&#xf1f8;',
				textsize='1.2',
				safe=True,
				type='label',
				order=49,
				require_confirm=False,
			)]

	async def delete_folder(self, player, values, folder_dictionary, view, **kwargs):
		# Check permission on folder.
		if folder_dictionary['type'] == 'auto':
			await show_alert(player, 'You can not remove an auto-folder!', size='sm')
			return

		# Check permission on folder.
		# Masteradmins (level 3) are allowed to delete other folders.
		if folder_dictionary['owner_login'] != player.login and player.level < 3:
			await show_alert(player, 'You can not remove folders made by others!', size='sm')
			return

		# Ask for confirmation.
		cancel = bool(await ask_confirmation(player, 'Are you sure you want to remove folder \'{}\'$z$s?'.format(
			folder_dictionary['name']
		), size='sm'))
		if cancel is True:
			return

		# Remove maps from folder.
		await MapInFolder.execute(
			MapInFolder.delete()
				.where(MapInFolder.folder == folder_dictionary['id'].replace('database_', ''))
		)

		# Remove folder.
		await MapFolder.execute(
			MapFolder.delete()
				.where(MapFolder.id == folder_dictionary['id'].replace('database_', ''))
		)

		# Refresh list.
		await self.refresh(player)

	async def action_show(self, player, values, instance, **kwargs):
		await self.folder_manager.display_folder(player, instance)

	async def create_folder(self, player, values, **kwargs):
		if self.child:
			return

		self.child = CreateFolderView(self, player, self.folder_manager)
		await self.child.display()
		await self.child.wait_for_response()
		await self.child.destroy()
		await self.display(self.player)  # refresh.
		self.child = None

	async def get_data(self):
		return await self.folder_manager.get_folders(self.player)


class CreateFolderView(TemplateView):
	"""
	View to create new folder.
	"""
	template_name = 'jukebox/folder_create.xml'

	def __init__(self, parent, player, folder_manager):
		"""
		Initiate child create view.

		:param parent: Parent view.
		:param player: Player instance.
		:param folder_manager: Folder manager instance.
		:type parent: pyplanet.view.base.View
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type folder_manager: pyplanet.apps.contrib.jukebox.folders.FolderManager
		"""
		super().__init__(parent.manager)

		self.parent = parent
		self.player = player
		self.folder_manager = folder_manager
		self.app = folder_manager.app

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)
		self.subscribe('button_save', self.save)
		self.subscribe('button_cancel', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()
		context['is_admin'] = self.player.level >= self.player.LEVEL_ADMIN
		return context

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])

		self.response_future.set_result(None)
		self.response_future.done()

	async def wait_for_response(self):
		return await self.response_future

	async def save(self, player, action, values, *args, **kwargs):
		"""
		Save action.

		:param player: Player instance
		:param action: Action label
		:param values: Values from manialink
		:param args: *
		:param kwargs: **
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""

		folder_name = values['folder_name']
		folder_privacy = values['folder_privacy']

		if folder_privacy == 'public' and player.level < player.LEVEL_ADMIN:
			folder_privacy = 'private'

		# Check if the user has already created 5 private folders
		current_folders = await self.folder_manager.get_private_folders(player)
		if folder_privacy == 'private' and len(current_folders) >= 5 and player.level < player.LEVEL_ADMIN:
			self.response_future.set_result(None)
			self.response_future.done()
			return await show_alert(player, 'You reached the maximum of 5 private folders!', 'sm')

		# Check if name is valid.
		if len(folder_name) < 3:
			self.response_future.set_result(None)
			self.response_future.done()
			return await show_alert(player, 'The name you gave is not valid. Please provide a name with at least 3 characters.', 'sm')

		# Create folder.
		await self.folder_manager.create_folder(name=folder_name, player=player, public=folder_privacy == 'public')

		# Return response.
		self.response_future.set_result(None)
		self.response_future.done()


class AddToFolderView(TemplateView):
	"""
	Add map to folder view.
	"""
	template_name = 'jukebox/folder_add.xml'

	def __init__(self, app, player, map_id, folder_manager):
		"""
		Initiate child add-to-folder view.

		:param app: App Instance.
		:param player: Player instance.
		:param map_id: Map database ID.
		:param folder_manager: Folder manager instance.
		:type app: pyplanet.apps.contrib.jukebox.Jukebox
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type folder_manager: pyplanet.apps.contrib.jukebox.folders.FolderManager
		"""
		super().__init__(app.context.ui)
		self.app = app
		self.player = player
		self.map_id = map_id
		self.folder_manager = folder_manager

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)
		self.subscribe('button_cancel', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()
		context['folders'] = (await self.folder_manager.get_writable_folders(self.player))[:20]
		return context

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])

		self.response_future.set_result(None)
		self.response_future.done()

	async def wait_for_response(self):
		return await self.response_future

	async def handle_catch_all(self, player, action, values, **kwargs):
		# Filter out the folder selection.
		if not action.startswith('folder_select__'):
			return

		# Fetch the folder from the database.
		try:
			folder = await MapFolder.get(id=action[15:])
		except:
			return  # Ignore exceptions here, could be that the folder has been deleted recently.

		# Check permission on folder.
		if folder.public and player.level < player.LEVEL_ADMIN:
			return
		if not folder.public and folder.player_id != player.id:
			return

		# Add map to folder.
		if not await self.folder_manager.add_map_to_folder(folder.id, self.map_id):
			await show_alert(player, 'Map already in folder!', 'sm')
			return

		self.response_future.set_result(True)
		self.response_future.done()
